#include "../index/regex.h"
#include "suffix_tree.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>

#define DIR "/mjolnir/thomas/research/display"
#define PATHFILE "/tmp/path_file"
#define SUBFILE "/tmp/sub_file"
#define SUPERFILE "/tmp/sup_file"

ofstream StatisticsFileStream;

static int experiment = 1;

// anchor is best explained with an example
// if path_expn is ?.+.tag.+.tag3.?, then anchor = tag
// if path_expn is ?.?.+, then anchor = ""
char* find_anchor(char *path_expn_) {
  char *token = strtok(path_expn_, ".?+");
  if(token) return strdup(token);
  else return strdup("");
}

// Args: path_expn_ : expression of the form (tagname | ? | +) ((.) (tagname | ? | +))*
//                    where ? means any tagname and + means 1 more tagnames
//       in_fname_  : file containing <path, offset, num_docs> records
//      out_fname_  : output file
//     root_super_  : output super tree
//       root_sub_  : output sub tree
void anchor_fn(char *path_expn_, char *in_fname_, char *out_fname_, 
	    suffix_tree_node *&root_super_, suffix_tree_node *&root_sub_) {
  int in_fd, out_fd, path_len, file_off, num_docs, init_seg_len, fina_seg_len, count=1, anchor_len;
  char *path_ptr, *path_expn_copy = strdup(path_expn_), *anchor, *anchor_ptr;
  char *anchored_sub_path, *anchored_super_path_reverse;
  char *sub_path_reverse, *super_path_reverse, *super_path, *sub_path, *sub_path_copy, *super_path_copy;

  struct timeval startTime;
  struct timeval stopTime;

  anchor = find_anchor(path_expn_copy);
  anchor_len = strlen(anchor);
  free(path_expn_copy);
  //cout << "anchor = " << anchor << endl;
  
  in_fd = open(in_fname_, O_RDONLY);
  out_fd = open(out_fname_, O_CREAT | O_WRONLY | O_TRUNC, S_IRWXU);

  int total_npaths = 0, match_npaths = 0; 
  double insert_time_path = 0, insert_time = 0, match_time_path = 0, match_time = 0;
  double read_time_path = 0, read_time = 0, write_time_path = 0, write_time;
  double split_time_path = 0, split_time = 0;

  while(1) {    
    assert(gettimeofday(&startTime, NULL) != -1);
    if(read(in_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) break;
    path_ptr = new char[path_len + 1];
    if(read(in_fd, (char *) path_ptr, path_len) < path_len) return;
    path_ptr[path_len] = 0;
    assert(gettimeofday(&stopTime, NULL) != -1);
    read_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
		       (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
    read_time += read_time_path;

    if(read(in_fd, (char *)&file_off, sizeof(int)) < sizeof(int)) return;
    if(read(in_fd, (char *)&num_docs, sizeof(int)) < sizeof(int)) return;
    total_npaths++;
    
    assert(gettimeofday(&startTime, NULL) != -1);
    // VLDB hack -- Thomas
    char *anchor_tag1 = new char[strlen(anchor)+2];
    sprintf(anchor_tag1, ".%s.", anchor);
    anchor_len = strlen(anchor_tag1);
    anchor_ptr = strstr(path_ptr, anchor_tag1);
    delete []anchor_tag1;
    // VLDB hack end

    if(anchor_ptr != NULL) {
      match_npaths++;
      assert(gettimeofday(&stopTime, NULL) != -1);
      match_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			 (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      match_time += match_time_path;

      assert(gettimeofday(&startTime, NULL) != -1);
      sub_path = anchor_ptr + anchor_len;
      if(sub_path[0] == '.') sub_path++;
      sub_path_copy = strdup(sub_path);
      sub_path_reverse = reverse_path(sub_path_copy);
      free(sub_path_copy);
      fina_seg_len = strlen(sub_path);
      
      init_seg_len = anchor_ptr - path_ptr;
      if(init_seg_len > 0 && path_ptr[init_seg_len-1] == '.') init_seg_len--;      
      super_path = new char[init_seg_len + 1];
      strncpy(super_path, path_ptr, init_seg_len);
      super_path[init_seg_len] = 0;
      super_path_copy = strdup(super_path);
      super_path_reverse = reverse_path(super_path_copy);
      free(super_path_copy);
      assert(gettimeofday(&stopTime, NULL) != -1);
      split_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			 (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      split_time += split_time_path;

      assert(gettimeofday(&startTime, NULL) != -1);
      if(write(out_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) return;
      if(write(out_fd, (char *) path_ptr, path_len) < path_len) return;
      assert(gettimeofday(&stopTime, NULL) != -1);
      write_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			 (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      write_time += write_time_path;

      if(write(out_fd, (char *)&file_off, sizeof(int)) < sizeof(int)) return;
      if(write(out_fd, (char *)&num_docs, sizeof(int)) < sizeof(int)) return;
      
      //cout << count << ". pat: " << path_ptr << endl;
      assert(gettimeofday(&startTime, NULL) != -1);
      if(fina_seg_len > 0) {	
	// VLDB hack -- Thomas
	anchored_sub_path = new char[fina_seg_len + anchor_len + 2];
	sprintf(anchored_sub_path, "%s.%s", anchor, sub_path);
	//cout << count << ". sub: " << anchor_ptr << " (" << super_path << ") " << endl;      
	// VLDB hack end
	// cout << count << ". sub: " << anchored_sub_path << " (" << super_path << ") " << endl;      
	insert_path(root_sub_, "", anchored_sub_path, super_path, file_off, num_docs);
      }
      if(init_seg_len > 0) {
	//anchor_len cannot be 0 if init_seg_len > 0
	anchored_super_path_reverse = new char[init_seg_len + anchor_len + 2];
	sprintf(anchored_super_path_reverse, "%s.%s", anchor, super_path_reverse);
	// cout << count << ". sup: " << anchored_super_path_reverse << " (" << sub_path_reverse <<  ") " << endl;
	insert_path(root_super_, "", anchored_super_path_reverse, sub_path_reverse, file_off, num_docs);
	delete []anchored_super_path_reverse;
      }
      assert(gettimeofday(&stopTime, NULL) != -1);
      
      insert_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			  (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      insert_time += insert_time_path;

      delete []super_path_reverse;
      delete []super_path;
      delete []sub_path_reverse;
      count++;
    }
    else {
      assert(gettimeofday(&stopTime, NULL) != -1);
      match_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			 (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      match_time += match_time_path;
    }
    delete []path_ptr;
  }
  if(experiment == 1) {
    StatisticsFileStream << total_npaths << " "; // 1. Total Num Paths
    StatisticsFileStream << match_npaths << " "; // 2. Matched Num Paths
    StatisticsFileStream << read_time << " "; // 3. Path Read Time
    StatisticsFileStream << match_time << " "; // 4. Regular Expression Match Time
    StatisticsFileStream << write_time << " "; // 5. Path Write Time
    StatisticsFileStream << split_time << " "; // 6. Path Split Time
    StatisticsFileStream << insert_time << " "; // 7. Suffix Tree Insertion Time
  }
  close(in_fd);
  close(out_fd);
  free(anchor);
  return;
}
int main(int argc, char *argv[]) {
  // open the statistics log file
  char *fname = new char[strlen(DIR) + strlen("/anchor_statistics.log") + 1];
  sprintf(fname, "%s/anchor_statistics.log", DIR);
  StatisticsFileStream.open(fname, ios::out | ios::app);
  StatisticsFileStream.setf(ios::fixed, ios::floatfield);
  delete []fname;

  struct timeval startTime;
  struct timeval stopTime;

  // get params
  int old_view = atoi(argv[1]);
  char *path_expn = argv[2];
  int new_view = atoi(argv[3]);
  int threshold = atoi(argv[4]);
  char *client_id = argv[5];

  // construct new path file, sub and super trees
  suffix_tree_node *root_super = NULL;
  suffix_tree_node *root_sub = NULL;

  char *path_file = new char[strlen(PATHFILE) + strlen(client_id) + 1];
  sprintf(path_file, "%s%s", PATHFILE, client_id);

  char *in_file = new char[strlen(path_file) + 10];
  char *out_file = new char[strlen(path_file) + 10];
  
  sprintf(in_file, "%s.%d", path_file, old_view);
  sprintf(out_file, "%s.%d", path_file, new_view);   

  anchor_fn(path_expn, in_file, out_file, root_super, root_sub);

  delete []in_file;
  delete []out_file;
  delete []path_file;

  // create new sub and super files
  char *sub_file = new char[strlen(SUBFILE) + strlen(client_id) + 1];
  sprintf(sub_file, "%s%s", SUBFILE, client_id);

  char *sup_file = new char[strlen(SUPERFILE) + strlen(client_id) + 1];
  sprintf(sup_file, "%s%s", SUPERFILE, client_id);

  char *out_sub_file = new char[strlen(sub_file) + 10];
  sprintf(out_sub_file, "%s.%d", sub_file, new_view);

  char *out_sup_file = new char[strlen(sup_file) + 10];
  sprintf(out_sup_file, "%s.%d", sup_file, new_view);

  double print_file_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  print_to_file(out_sub_file, root_sub);
  print_to_file(out_sup_file, root_super);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_file_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);

  if(experiment == 1) {
    StatisticsFileStream << print_file_time << " "; // 8. Suffix Tree Print To File Time
  }
  delete []out_sub_file;
  delete []out_sup_file;
  delete []sup_file;
  delete []sub_file;

  // print trees as HTML
  double print_html_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  int max_sup_level = 0, max_sub_level = 0;
  int sub_node_count = 0, sup_node_count = 0;
  // VLDB deadline -- Thomas, uncomment next 2 lines
  print_html_trees(new_view, threshold, root_super, root_sub, client_id, 
		   sub_node_count, max_sub_level, sup_node_count, max_sup_level);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_html_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  if(experiment == 1) {
    StatisticsFileStream << sub_node_count << " "; // 9. (for anchor) Node Count In Sub Tree
    StatisticsFileStream << max_sub_level << " "; // 10. (for anchor) Max Sub Tree Depth
    StatisticsFileStream << sup_node_count << " "; // 11. (for anchor) Node Count In Super Tree 
    StatisticsFileStream << max_sup_level << " "; // 12. (for anchor) Max Super Tree Depth 

    StatisticsFileStream << print_html_time << " "; // 13. Suffix Tree HTML Print Time
  }
  // delete trees
  double delete_trees_time;
  assert(gettimeofday(&startTime, NULL) != -1);
  delete_suffix_tree(root_sub);
  delete_suffix_tree(root_super);
  assert(gettimeofday(&stopTime, NULL) != -1);
  delete_trees_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		       (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  if(experiment == 1) {
    StatisticsFileStream << delete_trees_time << endl; // 14. Delete Trees Time
  }
  // cleanup
  StatisticsFileStream.close();

  return 0;
}

// 1. Total Num Paths
// 2. Matched Num Paths
// 3. Path Read Time
// 4. Regular Expression Match Time
// 5. Path Write Time
// 6. Path Split Time
// 7. Suffix Tree Insertion Time
// 8. Suffix Tree Print To File Time
// 9. (for anchor) Node Count In Sub Tree
// 10. (for anchor) Max Sub Tree Depth
// 11. (for anchor) Node Count In Super Tree
// 12. (for anchor) Max Super Tree Depth
// 13. Suffix Tree HTML Print Time
// 14. Delete Trees Time
