#include "getdocs.h"
#include "suffix_tree.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include "../index/common.h"
#include <fstream.h>
#include <assert.h>
#include "print_urls.h"

#define DIR "/mjolnir/thomas/research/display"
#define SUBFILE "/tmp/sub_file"
#define SUPERFILE "/tmp/sup_file"
#define DOCSFILE "/tmp/docs_file"
#define MERGEFILE "/tmp/merge_file"
#define URLFILE "/mjolnir5/thomas/index/index.ids"

#define RESULTSPERPAGE 10
#define MAX_DOCS_PER_BATCH 10
#define MAX_DOCS_PER_OUTPAGE 100

ofstream StatisticsFileStream;

void get_curr(merge_struct *table_, int slot_, int &doc_id_, int &rank_) {
  doc_id_ = table_[slot_].array[table_[slot_].position].doc_id;
  rank_ = table_[slot_].array[table_[slot_].position].rank;
}

// return 1 if slot_ has no more values remaining
int go_next(merge_struct *table_, int slot_, int docs_file_id_) {
  table_[slot_].position++;
  if(table_[slot_].position == table_[slot_].num_in_batch) {
    if(table_[slot_].num_docs_left == 0) return 1;
    else {
      table_[slot_].num_in_batch = (MAX_DOCS_PER_BATCH < table_[slot_].num_docs_left ? MAX_DOCS_PER_BATCH : table_[slot_].num_docs_left);
      table_[slot_].position = 0;
      if(lseek(docs_file_id_, table_[slot_].file_offset, SEEK_SET) < 0) return 1;
      for(int i=0; i<table_[slot_].num_in_batch; i++) {
	if(read(docs_file_id_, (char *)&table_[slot_].array[i].doc_id, sizeof(int)) < sizeof(int)) return 1;
	if(read(docs_file_id_, (char *)&table_[slot_].array[i].rank, sizeof(int)) < sizeof(int)) return 1;
      }
      table_[slot_].file_offset += (table_[slot_].num_in_batch * 2 * sizeof(int));
      table_[slot_].num_docs_left -= table_[slot_].num_in_batch;
      return 0;
    }
  }    
  else return 0;
}

int go_first(merge_struct *table_, int slot_, int docs_file_id_, int file_offset_, int num_docs_) {
  table_[slot_].file_offset = file_offset_;
  table_[slot_].num_docs_left = num_docs_;

  table_[slot_].num_in_batch = (MAX_DOCS_PER_BATCH < table_[slot_].num_docs_left ? MAX_DOCS_PER_BATCH : table_[slot_].num_docs_left);
  table_[slot_].position = 0;
  table_[slot_].array = new (doc_rank_struct)[MAX_DOCS_PER_BATCH];
  if(lseek(docs_file_id_, table_[slot_].file_offset, SEEK_SET) < 0) return 1;
  for(int i=0; i<table_[slot_].num_in_batch; i++) {
    if(read(docs_file_id_, (char *)&table_[slot_].array[i].doc_id, sizeof(int)) < sizeof(int)) return 1;
    if(read(docs_file_id_, (char *)&table_[slot_].array[i].rank, sizeof(int)) < sizeof(int)) return 1;
  }
  table_[slot_].file_offset += (table_[slot_].num_in_batch * 2 * sizeof(int));
  table_[slot_].num_docs_left -= table_[slot_].num_in_batch;
  if(table_[slot_].position == table_[slot_].num_in_batch) return 1;
  return 0;
}

void print_out_array(int size_, doc_rank_struct *file_ptr_, int qr_id) {
  for(int i=0; i<size_; i++) {
    if(write(qr_id, (char *)&(file_ptr_[i].doc_id), sizeof(int)) < sizeof(int)) return;
    if(write(qr_id, (char *)&(file_ptr_[i].rank), sizeof(int)) < sizeof(int)) return;    
  }
}

void insert_file_num(int &size_, file_doc_pair* &file_ptr_, int file_off_, int num_docs_) {
  size_++;
  file_doc_pair* temp = file_ptr_;
  file_ptr_ = new (file_doc_pair) [size_];
  for(int i=0; i<(size_-1); i++)
    file_ptr_[i] = temp[i];
  file_ptr_[size_-1].file_offset = file_off_;
  file_ptr_[size_-1].num_docs = num_docs_;
  delete []temp;
}

void getdocs_help(int st_file_id_, int &size_, file_doc_pair *& file_num_list_) {
  int bpath_len, is_leaf, label_len, super_path_len;
  int num_paths, file_off, num_docs, num_children;
  char *bpath, *super_path, *label;

  if(read(st_file_id_, (char *)&bpath_len, sizeof(int)) < sizeof(int)) return;
  bpath = new char[bpath_len + 1];
  if(read(st_file_id_, (char *) bpath, bpath_len) < bpath_len) return;
  bpath[bpath_len] = 0;
  
  if(read(st_file_id_, (char *)&is_leaf, sizeof(int)) < sizeof(int)) return;
  if(is_leaf) {
    if(read(st_file_id_, (char *)&num_paths, sizeof(int)) < sizeof(int)) return;
    for(int i=0; i<num_paths; i++) {
      if(read(st_file_id_, (char *)&super_path_len, sizeof(int)) < sizeof(int)) return;
      super_path = new char[super_path_len + 1];
      if(read(st_file_id_, (char *)super_path, super_path_len) < super_path_len) return;
      super_path[super_path_len] = 0;
      if(read(st_file_id_, (char *)&file_off, sizeof(int)) < sizeof(int)) return;
      if(read(st_file_id_, (char *)&num_docs, sizeof(int)) < sizeof(int)) return;
      
      if(super_path_len > 0) {
	//path_len = super_path_len + 1 + bpath_len;
	//path_ptr = new char[path_len + 1];
	//sprintf(path_ptr, "%s.%s", super_path, bpath);	
	//if(write(path_file_id_, (char *)&path_len, sizeof(int)) < sizeof(int)) return;
	//if(write(path_file_id_, (char *) path_ptr, path_len) < path_len) return;
	//delete []path_ptr;
      }
      else { 
	//if(write(path_file_id_, (char *)&bpath_len, sizeof(int)) < sizeof(int)) return;
	//if(write(path_file_id_, (char *) bpath, bpath_len) < bpath_len) return;
      }

      insert_file_num(size_, file_num_list_, file_off, num_docs);
      delete []super_path;
    }
  }
  else {
    if(read(st_file_id_, (char *)&num_children, sizeof(int)) < sizeof(int)) return;
    for(int i=0; i<num_children; i++) {
      if(read(st_file_id_, (char *)&label_len, sizeof(int)) < sizeof(int)) return;
      label = new char[label_len + 1];
      if(read(st_file_id_, (char *)label, label_len) < label_len) return;
      label[label_len] = 0;
      delete []label;
      getdocs_help(st_file_id_, size_, file_num_list_);
    }
  }
  delete []bpath;
}


void getdocs(int node_id_, char *st_file_, int &size_, file_doc_pair *& file_num_list_) {
  int st_index_file_id, st_file_id, offset1, offset2, label_len;
  char *label;

  char *st_index_file = new char[strlen(st_file_) + strlen(".idx") + 1];
  sprintf(st_index_file, "%s.idx", st_file_);
  st_index_file_id = open(st_index_file, O_RDONLY);
  delete []st_index_file;

  if(lseek(st_index_file_id, node_id_ * 2 * sizeof(int), SEEK_SET) < 0) return;
  if(read(st_index_file_id, (char *)&offset1, sizeof(int)) < sizeof(int)) return;
  if(read(st_index_file_id, (char *)&offset2, sizeof(int)) < sizeof(int)) return;
  close(st_index_file_id);
  
  st_file_id = open(st_file_, O_RDONLY);
   
  if(lseek(st_file_id, offset2, SEEK_SET) < 0) return;
  if(read(st_file_id, (char *)&label_len, sizeof(int)) < sizeof(int)) return;
  label = new char[label_len + 1];
  if(read(st_file_id, (char *)label, label_len) < label_len) return;
  label[label_len] = 0;
  delete []label;

  size_ = 0;
  file_num_list_ = NULL;
  getdocs_help(st_file_id,  size_, file_num_list_);
  
  close(st_file_id);
}

void merge(int num_lists_, file_doc_pair* file_ptr_, char *docs_file_, 
	   char *merge_file_, int &out_docs_count_, int &phrase_count_) {
  int min_doc_id, new_rank, doc_id, old_rank, skip_count;
  int same_doc_count, done_slot_count, i;
  doc_rank_struct* out_array = new (doc_rank_struct)[MAX_DOCS_PER_OUTPAGE];
  int docs_file_id = open(docs_file_, O_RDONLY);
  merge_struct* table = new (merge_struct) [num_lists_];
  int *done_slots = new (int) [num_lists_];
  int *same_doc_slots = new (int) [num_lists_];
  int qr_id = open(merge_file_, O_CREAT | O_WRONLY | O_TRUNC, S_IWRITE | S_IREAD);

  if(read(docs_file_id, (char *)&phrase_count_, sizeof(int)) < sizeof(int)) return;

  done_slot_count = 0;
  for(i = 0; i<num_lists_; i++) {
    done_slots[i] = go_first(table, i, docs_file_id, file_ptr_[i].file_offset, file_ptr_[i].num_docs);
    if(done_slots[i] == 1) done_slot_count++;
  }

  out_docs_count_ = 0;
  while(done_slot_count < num_lists_) {
    same_doc_count = 0;
    skip_count = 0;
    for(i=0; i<num_lists_; i++) {
      if(done_slots[i] == 1) {
	skip_count++;
	continue;
      }

      get_curr(table, i, doc_id, old_rank);
      if(skip_count == i || doc_id < min_doc_id) {
	min_doc_id = doc_id;
	new_rank = old_rank;
	same_doc_slots[0] = i;
	same_doc_count = 1;	
      }
      else if(doc_id == min_doc_id) {
	new_rank = old_rank | new_rank;	
	same_doc_slots[same_doc_count] = i;
	same_doc_count++;
      }
    }
    
    for(i=0; i<same_doc_count; i++) {
      if(go_next(table, same_doc_slots[i], docs_file_id)) {
	if(done_slots[same_doc_slots[i]] == 0) {
	  done_slots[same_doc_slots[i]] = 1;
	  done_slot_count++;
	}
      }
    }
    // output min_doc_id && new_rank
    out_array[out_docs_count_%MAX_DOCS_PER_OUTPAGE].doc_id = min_doc_id;
    out_array[out_docs_count_%MAX_DOCS_PER_OUTPAGE].rank  = new_rank;
    out_docs_count_++;
    if(out_docs_count_%MAX_DOCS_PER_OUTPAGE == 0) {
      // print the out array
      print_out_array(MAX_DOCS_PER_OUTPAGE, out_array, qr_id);
    }
  }

  if(out_docs_count_%MAX_DOCS_PER_OUTPAGE > 0) {
    // print the out array
    print_out_array(out_docs_count_%MAX_DOCS_PER_OUTPAGE, out_array, qr_id);
  }

  for(i=0; i<num_lists_; i++) 
    delete [](table[i].array);
  delete []table;
  delete []out_array;
  delete []done_slots;
  delete []same_doc_slots;
  close(docs_file_id);
  close(qr_id);
  return;
}



int main(int argc, char *argv[]) {
  char *fname = new char[strlen(DIR) + strlen("/getdocs_statistics.log") + 1];
  sprintf(fname, "%s/getdocs_statistics.log", DIR);
  StatisticsFileStream.open(fname, ios::out | ios::app);
  StatisticsFileStream.setf(ios::fixed, ios::floatfield);
  delete []fname;

  char *in_file;

  int old_view = atoi(argv[1]); 
  int leftOrRight = atoi(argv[2]); // 0 => left, 1 => right
  int node_id = atoi(argv[3]); // starting from 0
  int start = atoi(argv[4]);
  char *client_id = argv[5];
  int num_docs;

  char *sub_file = new char[strlen(SUBFILE) + strlen(client_id) + 1];
  sprintf(sub_file, "%s%s", SUBFILE, client_id);

  char *sup_file = new char[strlen(SUPERFILE) + strlen(client_id) + 1];
  sprintf(sup_file, "%s%s", SUPERFILE, client_id);

  char *docs_file = new char[strlen(DOCSFILE) + strlen(client_id) + 1];
  sprintf(docs_file, "%s%s", DOCSFILE, client_id);

  char *merge_file = new char[strlen(MERGEFILE) + strlen(client_id) + 1];
  sprintf(merge_file, "%s%s", MERGEFILE, client_id);

  char *url_file = URLFILE;

  file_doc_pair* file_num_list = NULL;
  int size = 0;
  
  if(leftOrRight == 1) {
    in_file = new char[strlen(sub_file) + 10];
    sprintf(in_file, "%s.%d", sub_file, old_view);
  }
  else {
    in_file = new char[strlen(sup_file) + 10];
    sprintf(in_file, "%s.%d", sup_file, old_view);
  }

  getdocs(node_id, in_file, size, file_num_list);
  //cout << "----------------" << endl;
  //for(int i=0; i<size; i++) {
  //cout << file_num_list[i].file_offset<< " " << file_num_list[i].num_docs << endl;
  //}
  //cout << "----------------" << endl;
  int phrase_count = 0;
  merge(size, file_num_list, docs_file, merge_file, num_docs, phrase_count);

  int batch_size = RESULTSPERPAGE;
  print_html_urls(merge_file, start, batch_size, num_docs, url_file, phrase_count, client_id);

  delete []file_num_list;
  delete []in_file;
  delete [] sub_file;
  delete [] sup_file;
  delete [] docs_file;
  delete [] merge_file;
  StatisticsFileStream.close();
}

