#include "../index/regex.h"
#include "suffix_tree.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>

#define DIR "/mjolnir/thomas/research/display"
#define PATHFILE "/tmp/path_file"
#define SUBFILE "/tmp/sub_file"

ofstream StatisticsFileStream;

static int experiment = 1; 
// set this to 1 to run any experiment
// output of experiments is written to StatisticsFileStream

//This block of code is for regular expression support
#define INIT         register char *sp=instring;
#define GETC()       (*sp++)
#define PEEKC()      (*sp)
#define UNGETC(c)     (--sp)
#define RETURN(c)    return 0;
#define ERROR(c)     exit(0);
#include <regexp.h>

char *compile(char *instring, char *expbuf,
	      const char *endbuf, int eof);

int step(const char *string, const char *expbuf);

int advance(const char *string, const char *expbuf);

extern char *loc1, *loc2, *locs;

// Args: path_expn_ : expression of the form (tagname | ? | +) ((.) (tagname | ? | +))*
//                    where ? means any tagname and + means 1 more tagnames
//       in_fname_  : file containing <path, offset, num_docs> records
//      out_fname_  : output file
//       root_sub_  : output sub tree
void refine(char *path_expn_, char *in_fname_, char *out_fname_, 
	    suffix_tree_node *&root_sub_) {
  char regString[ESIZE];
  char expbuf[ESIZE];
  int in_fd, out_fd, path_len, file_off, num_docs, count=1;
  char *path_ptr;

  struct timeval startTime;
  struct timeval stopTime;

  if(createRegExp(path_expn_, regString, ESIZE) == -1) return;
  compile((char *)regString, (char *)expbuf, (const char *)(&expbuf[ESIZE]), '\0');
  //cout << "regexp = " << regString << endl;

  in_fd = open(in_fname_, O_RDONLY);
  out_fd = open(out_fname_, O_CREAT | O_WRONLY | O_TRUNC, S_IRWXU);

  int total_npaths = 0, match_npaths = 0; 
  double insert_time_path = 0, insert_time = 0, match_time_path = 0, match_time = 0;
  double read_time_path = 0, read_time = 0, write_time_path = 0, write_time;

  while(1) {    
    assert(gettimeofday(&startTime, NULL) != -1);
    if(read(in_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) break;
    path_ptr = new char[path_len + 1];
    if(read(in_fd, (char *) path_ptr, path_len) < path_len) return;
    path_ptr[path_len] = 0;
    assert(gettimeofday(&stopTime, NULL) != -1);
    read_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
		       (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
    read_time += read_time_path;

    if(read(in_fd, (char *)&file_off, sizeof(int)) < sizeof(int)) return;
    if(read(in_fd, (char *)&num_docs, sizeof(int)) < sizeof(int)) return;
    total_npaths++;
    
    assert(gettimeofday(&startTime, NULL) != -1);
    if(step((const char *)path_ptr, (const char *)expbuf)) {
      match_npaths++;
      assert(gettimeofday(&stopTime, NULL) != -1);
      match_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			 (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      match_time += match_time_path;

      assert(gettimeofday(&startTime, NULL) != -1);
      if(write(out_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) return;
      if(write(out_fd, (char *) path_ptr, path_len) < path_len) return;
      assert(gettimeofday(&stopTime, NULL) != -1);
      write_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			 (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      write_time += write_time_path;

      if(write(out_fd, (char *)&file_off, sizeof(int)) < sizeof(int)) return;
      if(write(out_fd, (char *)&num_docs, sizeof(int)) < sizeof(int)) return;
      
      //cout << count << ". pat: " << path_ptr << endl;
      assert(gettimeofday(&startTime, NULL) != -1);
      insert_path(root_sub_, "", path_ptr, "", file_off, num_docs);
      assert(gettimeofday(&stopTime, NULL) != -1);
      
      insert_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			  (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      insert_time += insert_time_path;

      count++;
    }
    else {
      assert(gettimeofday(&stopTime, NULL) != -1);
      match_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			 (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      match_time += match_time_path;
    }
    delete []path_ptr;
  }
  if(experiment == 1) {
    StatisticsFileStream << total_npaths << " "; // 1. Total Num Paths
    StatisticsFileStream << match_npaths << " "; // 2. Matched Num Paths
    StatisticsFileStream << read_time << " "; // 3. Path Read Time
    StatisticsFileStream << match_time << " "; // 4. Regular Expression Match Time
    StatisticsFileStream << write_time << " "; // 5. Path Write Time
    StatisticsFileStream << insert_time << " "; // 6. Suffix Tree Insertion Time
  }
  close(in_fd);
  close(out_fd);
  return;
}
int main(int argc, char *argv[]) {
  // open the statistics log file
  char *fname = new char[strlen(DIR) + strlen("/refine_statistics.log") + 1];
  sprintf(fname, "%s/refine_statistics.log", DIR);
  StatisticsFileStream.open(fname, ios::out | ios::app);
  StatisticsFileStream.setf(ios::fixed, ios::floatfield);
  delete []fname;

  struct timeval startTime;
  struct timeval stopTime;

  // get params
  int old_view = atoi(argv[1]);
  char *path_expn = argv[2];
  int new_view = atoi(argv[3]);
  int threshold = atoi(argv[4]);
  char *client_id = argv[5];

  // construct new path file, sub tree
  suffix_tree_node *root_sub = NULL;

  char *path_file = new char[strlen(PATHFILE) + strlen(client_id) + 1];
  sprintf(path_file, "%s%s", PATHFILE, client_id);

  char *in_file = new char[strlen(path_file) + 10];
  char *out_file = new char[strlen(path_file) + 10];
  
  sprintf(in_file, "%s.%d", path_file, old_view);
  sprintf(out_file, "%s.%d", path_file, new_view);   

  refine(path_expn, in_file, out_file, root_sub);

  delete []in_file;
  delete []out_file;
  delete []path_file;

  // create new sub files
  char *sub_file = new char[strlen(SUBFILE) + strlen(client_id) + 1];
  sprintf(sub_file, "%s%s", SUBFILE, client_id);

  char *out_sub_file = new char[strlen(sub_file) + 10];
  sprintf(out_sub_file, "%s.%d", sub_file, new_view);

  double print_file_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  print_to_file(out_sub_file, root_sub);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_file_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  
  if(experiment == 1) {
    StatisticsFileStream << print_file_time << " "; // 7. Suffix Tree Print To File Time
  }
  delete []out_sub_file;
  delete []sub_file;

  // print trees as HTML
  double print_html_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  int max_sup_level = 0, max_sub_level = 0;
  int sub_node_count = 0, sup_node_count = 0;
  // VLDB deadline -- Thomas, Uncomment next 2 lines
  print_html_trees(new_view, threshold, NULL, root_sub, client_id, 
		   sub_node_count, max_sub_level, sup_node_count, max_sup_level);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_html_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  if(experiment == 1) {
    StatisticsFileStream << sub_node_count << " "; // 8. (for refine) Node Count In Sub Tree
    StatisticsFileStream << max_sub_level << " "; // 9. (for refine) Max Sub Tree Depth
    StatisticsFileStream << sup_node_count << " "; // 10. (for refine) Node Count In Super Tree 
    StatisticsFileStream << max_sup_level << " "; // 11. (for refine) Max Super Tree Depth 
    StatisticsFileStream << print_html_time << " "; // 12. Suffix Tree HTML Print Time
  }
  // delete trees
  double delete_trees_time;
  assert(gettimeofday(&startTime, NULL) != -1);
  delete_suffix_tree(root_sub);
  assert(gettimeofday(&stopTime, NULL) != -1);
  delete_trees_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		       (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  if(experiment == 1) {
    StatisticsFileStream << delete_trees_time << endl; // 14. Delete Trees Time
  }

  // cleanup
  StatisticsFileStream.close();

  return 0;
}

// 1. Total Num Paths
// 2. Matched Num Paths
// 3. Path Read Time
// 4. Regular Expression Match Time
// 5. Path Write Time
// 6. Suffix Tree Insertion Time
// 7. Suffix Tree Print To File Time
// 8. (for refine) Node Count In Sub Tree
// 9. (for refine) Max Sub Tree Depth
// 10. (for refine) Node Count In Super Tree
// 11. (for refine) Max Super Tree Depth
// 12. Suffix Tree HTML Print Time
// 13. Delete Trees Time
