#ifndef SUFFIX_TREE_H
#define SUFFIX_TREE_H

typedef struct suffix_tree_node_ {
  int doc_count;
  int level_num;
  char *path;
  bool is_leaf;
  union {
    struct {
      int num_children;  
      char** labels;
      struct suffix_tree_node_** children;
      int* num_docs;
    } non_leaf;
    struct {
      int num_paths;
      char** super_paths;
      int* file_offs;
      int* num_docs;
    } leaf;
  } u;
} suffix_tree_node;

typedef struct tree_node_info_ {
  char *label;
  int level;
  int is_leaf;
  int num_docs;
  int node_id;
} tree_node;

char *reverse_path(char *path_);

void insert_path(suffix_tree_node *&root_, char *path_, char *sub_path_, char *super_path_, int file_off_, int num_docs_);
void delete_suffix_tree(suffix_tree_node *&root_);
void print_text(suffix_tree_node *root_);
void print_html(int database_, int threshold_, suffix_tree_node *root_);
void print_to_file(char *st_file_, suffix_tree_node *root_);
void read_from_file(char *st_file_, suffix_tree_node *&root_);
void zoom(int node_id, char *st_file_, char *path_file_, 
	  suffix_tree_node *&root_same_, suffix_tree_node *&root_other_);
void print_html_trees(int view_, int threshold_, 
		      suffix_tree_node *root_sup_, suffix_tree_node *root_sub_, char *client_id,
		      int &sub_node_count, int &max_sub_level, 
		      int &sup_node_count, int &max_sup_level);
void linearize_tree(int level_, int print_flag_, int threshold_, int &node_id_, suffix_tree_node *root_, 
		    int &max_level_, tree_node* &tree_node_list_, int &count_);
void get_tree_stats(int level_, int print_flag_, int threshold_, int &node_id_, suffix_tree_node *root_, 
		    int &max_level_, int &count_, int &size_);
#endif
