#include "suffix_tree.h"
#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <assert.h>
#include <sys/time.h>

#define DIR "/mjolnir/thomas/research/display"
#define PATHFILE "/tmp/path_file"
#define SUBFILE "/tmp/sub_file"
#define SUPERFILE "/tmp/sup_file"

ofstream StatisticsFileStream;

// args: old_view, left-or-right (0=left, 1=right), nodeid, new_view, clientid
// sample call: ./zoom 1 1 1 2 0000
// init_view is initialized to 1 in find_inv_lists.cc
int main(int argc, char *argv[]) {
  // open stats log
  char *fname = new char[strlen(DIR) + strlen("/zoom_statistics.log") + 1];
  sprintf(fname, "%s/zoom_statistics.log", DIR);
  StatisticsFileStream.open(fname, ios::out | ios::app);
  StatisticsFileStream.setf(ios::fixed, ios::floatfield);
  delete []fname;

  struct timeval startTime;
  struct timeval stopTime;

  suffix_tree_node *root_sub, *root_sup;
  char *in_file;

  // get params
  int old_view = atoi(argv[1]); 
  int leftOrRight = atoi(argv[2]); // 0 => left, 1 => right
  int node_id = atoi(argv[3]); // starting from 0
  int new_view = atoi(argv[4]);
  char *client_id = argv[5];

  char *sub_file = new char[strlen(SUBFILE) + strlen(client_id) + 1];
  sprintf(sub_file, "%s%s", SUBFILE, client_id);

  char *sup_file = new char[strlen(SUPERFILE) + strlen(client_id) + 1];
  sprintf(sup_file, "%s%s", SUPERFILE, client_id);

  char *pat_file = new char[strlen(PATHFILE) + strlen(client_id) + 1];
  sprintf(pat_file, "%s%s", PATHFILE, client_id);


  if(leftOrRight == 1) {
    in_file = new char[strlen(sub_file) + 10];
    sprintf(in_file, "%s.%d", sub_file, old_view);
  }
  else {
    in_file = new char[strlen(sup_file) + 10];
    sprintf(in_file, "%s.%d", sup_file, old_view);
  }
  char *out_file = new char[strlen(pat_file) + 10];
  sprintf(out_file, "%s.%d", pat_file, new_view);

  root_sub = NULL;
  root_sup = NULL;
  if(leftOrRight == 1) 
    zoom(node_id, in_file, out_file, root_sub, root_sup);
  else
    zoom(node_id, in_file, out_file, root_sup, root_sub);

  char *out_sub_file = new char[strlen(sub_file) + 10];
  sprintf(out_sub_file, "%s.%d", sub_file, new_view);

  char *out_sup_file = new char[strlen(sup_file) + 10];
  sprintf(out_sup_file, "%s.%d", sup_file, new_view);

  double print_file_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  print_to_file(out_sub_file, root_sub);
  print_to_file(out_sup_file, root_sup);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_file_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  StatisticsFileStream << print_file_time << " "; // 3. Suffix Tree Print To File Time

  int max_sup_level = 0, max_sub_level = 0;
  int sub_node_count = 0, sup_node_count = 0;
  double print_html_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  // SIGMOD deadline: fast break in code
  // uncomment the next line before launching system - Thomas
  //print_html_trees(new_view, 0, root_sup, root_sub, client_id, sub_node_count, max_sub_level, sup_node_count, max_sup_level);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_html_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  StatisticsFileStream << sub_node_count << " "; // 4. Node Count In Sub Tree
  StatisticsFileStream << max_sub_level << " "; // 5. Max Sub Tree Depth
  StatisticsFileStream << sup_node_count << " "; // 6. Node Count In Super Tree 
  StatisticsFileStream << max_sup_level << " "; // 7. Max Super Tree Depth 
  StatisticsFileStream << print_html_time << " "; // 8. Suffix Tree HTML Print Time
  
  delete []in_file;
  delete []out_file;
  delete []out_sub_file;
  delete []out_sup_file;

  // delete trees
  double delete_trees_time;
  assert(gettimeofday(&startTime, NULL) != -1);
  if(root_sub) delete_suffix_tree(root_sub);
  if(root_sup) delete_suffix_tree(root_sup);
  assert(gettimeofday(&stopTime, NULL) != -1);
  delete_trees_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		       (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  StatisticsFileStream << delete_trees_time << endl; // 9. delete trees time

  StatisticsFileStream.close();
  delete []sub_file;
  delete []sup_file;
  delete []pat_file;
  return 0;
}
// 1. Visited Node Count
// 2. Zoom Time
// 3. Suffix Tree Print To File Time
// 4. Node Count In Sub Tree
// 5. Max Sub Tree Depth
// 6. Node Count In Super Tree 
// 7. Max Super Tree Depth 
// 8. Suffix Tree HTML Print Time
// 9. delete trees time
