#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "../index/LinearHash3.h"
#include "../index/String.h"
#include "../index/key_offset_record.h"
#include "../index/Int.h"
#include "../index/InvertedFile5.h"
#include "../display/suffix_tree.h"
#include "../merge/merge_inv_files.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <iostream.h>
#include <fstream.h>

ofstream GLOBAL_logFileStream;
char *log_dir, *tmp_dir;

#define BATCH_SIZE 10000 // num tuples
#define MAX_DOCS_PER_BATCH 2000
#define MAX_DOCS_PER_OUTPAGE 2000
#define MAX_PHRASE_LEN 200

extern "C" int errno;
ofstream StatisticsFileStream;
ofstream DocsStatisticsFileStream;
int experiment = 0; 

// = 1 :focus on the or function
// = -1: focus on the suffix tree print
// = 0 : focus on the suffix tree construction
int find_inv_lists(char *key_, char *fname_, int &offset_, bool &found_, double &dict_lookup_time_) {
  char *id_string, *out_file;
  LinearHash<String, Int, key_offset_record> *index;
  int index_num;
  Int offset_obj;
  struct timeval startTime;
  struct timeval stopTime;
  double find_offset_time;
  char *fname = new char[strlen(log_dir) + strlen("/find_inv_lists.log") + 1];
  sprintf(fname, "%s/find_inv_lists.log", log_dir);
  GLOBAL_logFileStream.open(fname);
  delete []fname;

  index_num = 0;
  if(strcmp(key_, "g") >= 0) index_num++;
  if(strcmp(key_, "n") >= 0) index_num++;
  if(strcmp(key_, "s") >= 0) index_num++;

  out_file = new char[strlen(fname_) + strlen(".idx.") + 10 + 1];
  sprintf(out_file, "%s.idx.%d", fname_, index_num);
  id_string = new char[strlen("Final.LH.") + 10 + 1];
  sprintf(id_string, "Final.LH.%d", index_num);
  
  assert(gettimeofday(&startTime, NULL) != -1);
  index = new LinearHash<String, Int, key_offset_record>(out_file, 0, 0, sizeof(int) + String::get_max_len(), id_string);
  // the number of bkts and slots_per_bkt are obtained from file header info
  // outfile has to exist for this to work
  
  found_ = false;
  if(index->find(String(key_), offset_obj, found_) == -1) exit(0);
  if(found_) offset_ = offset_obj.get_key();
  else offset_ = 0;
  assert(gettimeofday(&stopTime, NULL) != -1);
  find_offset_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		      (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  dict_lookup_time_ = find_offset_time;

  //StatisticsFileStream << "Find Offset Time=" << find_offset_time << endl;
    
  delete []out_file;
  delete []id_string;
  delete index;
  GLOBAL_logFileStream.close();
  return 0;
}


void get_curr(merge_struct *table_, int slot_, int &doc_id_, int &word_pos_) {
  doc_id_ = table_[slot_].array[table_[slot_].position].doc_id;
  word_pos_ = table_[slot_].array[table_[slot_].position].word_pos;
}

// return 1 if slot_ has no more values remaining
int go_next(merge_struct *table_, int slot_) {
  table_[slot_].position++;
  if(table_[slot_].position == table_[slot_].num_in_batch) {
    if(table_[slot_].num_tuples_left == 0) return 1;
    else {
      table_[slot_].num_in_batch = (MAX_DOCS_PER_BATCH < table_[slot_].num_tuples_left ? MAX_DOCS_PER_BATCH : table_[slot_].num_tuples_left);
      table_[slot_].position = 0;
      if(lseek(table_[slot_].file_fd, table_[slot_].file_offset, SEEK_SET) < 0) return 1;
      for(int i=0; i<table_[slot_].num_in_batch; i++) {
        if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].doc_id, sizeof(int)) < sizeof(int)) return 1;
        if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].word_pos, sizeof(int)) < sizeof(int)) return 1;
      }
      table_[slot_].file_offset += (table_[slot_].num_in_batch * 2 * sizeof(int));
      table_[slot_].num_tuples_left -= table_[slot_].num_in_batch;
      return 0;
    }
  }    
  else return 0;
}

int go_first(merge_struct *table_, int slot_, int file_fd_, int file_offset_, int num_tuples_) {
  table_[slot_].file_fd = file_fd_;
  table_[slot_].file_offset = file_offset_;
  table_[slot_].num_tuples_left = num_tuples_;

  table_[slot_].num_in_batch = (MAX_DOCS_PER_BATCH < table_[slot_].num_tuples_left ? 
				MAX_DOCS_PER_BATCH : table_[slot_].num_tuples_left);
  table_[slot_].position = 0;
  table_[slot_].array = new (doc_word_pos_struct)[MAX_DOCS_PER_BATCH];
  if(lseek(table_[slot_].file_fd, table_[slot_].file_offset, SEEK_SET) < 0) return 1;
  for(int i=0; i<table_[slot_].num_in_batch; i++) {
    if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].doc_id, sizeof(int)) < sizeof(int)) return 1;
    if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].word_pos, sizeof(int)) < sizeof(int)) return 1;
  }
  table_[slot_].file_offset += (table_[slot_].num_in_batch * 2 * sizeof(int));
  table_[slot_].num_tuples_left -= table_[slot_].num_in_batch;
  if(table_[slot_].position == table_[slot_].num_in_batch) return 1;
  return 0;
}

void print_out_array(int size_, doc_word_pos_struct *file_ptr_, int qr_id) {
  
  for(int i=0; i<size_; i++) {
    if(write(qr_id, (char *)&(file_ptr_[i].doc_id), sizeof(int)) < sizeof(int)) return;
    if(write(qr_id, (char *)&(file_ptr_[i].word_pos), sizeof(int)) < sizeof(int)) return;    
  }
  
  //for(int i=0; i<size_; i++) {
  //cout << file_ptr_[i].doc_id << " " << file_ptr_[i].word_pos << endl;
  //}
}

void merge_adj_lists(FileDescPtr first_ptr_, FileDescPtr second_ptr_, FileDescPtr result_ptr_, int distance_) {
  int first_doc_id, first_word_pos, second_doc_id, second_word_pos;
  int done_slot_count, i;
  doc_word_pos_struct* out_array = new (doc_word_pos_struct)[MAX_DOCS_PER_OUTPAGE];
  int num_lists = 2, out_docs_count;
  merge_struct* table = new (merge_struct) [num_lists];
  int *done_slots = new (int) [num_lists];

  lseek(result_ptr_.fd, result_ptr_.offset, SEEK_SET);

  done_slot_count = 0;
  done_slots[0] = go_first(table, 0, first_ptr_.fd, first_ptr_.offset, first_ptr_.ntuples);
  if(done_slots[0] == 1) done_slot_count++;
  done_slots[1] = go_first(table, 1, second_ptr_.fd, second_ptr_.offset, second_ptr_.ntuples);
  if(done_slots[1] == 1) done_slot_count++;  

  out_docs_count = 0;
  while(done_slot_count < num_lists) {
    if(!done_slots[0]) get_curr(table, 0, first_doc_id, first_word_pos);
    if(!done_slots[1]) get_curr(table, 1, second_doc_id, second_word_pos);

    if(!done_slots[0] && !done_slots[1]) {
      if(first_doc_id < second_doc_id || (first_doc_id == second_doc_id && first_word_pos < (second_word_pos - distance_))) {
	if(go_next(table, 0)) {
	  if(done_slots[0] == 0) { 
	    done_slots[0] = 1;
	    done_slot_count++;
	  }
	}           
      }
      
      if(first_doc_id > second_doc_id || (first_doc_id == second_doc_id && first_word_pos > (second_word_pos - distance_))) { 
	if(go_next(table, 1)) {
	  if(done_slots[1] == 0) { 
	    done_slots[1] = 1;
	    done_slot_count++;
	  }
	}           
      }
      
      if(first_doc_id == second_doc_id && first_word_pos ==  (second_word_pos - distance_)) {
	out_array[out_docs_count%MAX_DOCS_PER_OUTPAGE].doc_id = first_doc_id;
	out_array[out_docs_count%MAX_DOCS_PER_OUTPAGE].word_pos  = first_word_pos;
	out_docs_count++;
	if(out_docs_count%MAX_DOCS_PER_OUTPAGE == 0) {
	  // print the out array
	  print_out_array(MAX_DOCS_PER_OUTPAGE, out_array, result_ptr_.fd);
	}
	
	if(go_next(table, 0)) {
	  if(done_slots[0] == 0) { 
	    done_slots[0] = 1;
	    done_slot_count++;
	  }
	}           
	if(go_next(table, 1)) {
	  if(done_slots[1] == 0) { 
	    done_slots[1] = 1;
	    done_slot_count++;
	  }
	}      
      }
    }
    else if(!done_slots[0]) {
      if(go_next(table, 0)) {
	if(done_slots[0] == 0) { 
	    done_slots[0] = 1;
	    done_slot_count++;
	}
      }                 
    } 
    else if(!done_slots[1]) {
      if(go_next(table, 1)) {
	if(done_slots[1] == 0) { 
	  done_slots[1] = 1;
	  done_slot_count++;
	}
      }
    }
  }

  if(out_docs_count%MAX_DOCS_PER_OUTPAGE > 0) {
    // print the out array
    print_out_array(out_docs_count%MAX_DOCS_PER_OUTPAGE, out_array, result_ptr_.fd);
  }
  
  for(i=0; i<num_lists; i++) 
    delete [](table[i].array);
  delete []table;
  delete []out_array;
  delete []done_slots;
  return;
}

void merge_adj_inv_lists(File_Ptr &first_, File_Ptr &second_, File_Ptr &result_, int distance_) {
  int first_fd, second_fd, result_fd, i, j, condition;
  int first_npaths, second_npaths, result_npaths, end_offset, first_path_len, second_path_len;
  int first_ntuples, second_ntuples, num_tuples, result_ntuples, record_end_offset, num_tuples_offset;  
  bool read_first, read_second;
  char *first_buffer, *second_buffer, *first_path, *second_path;
  FileDescPtr first_ptr, second_ptr, result_ptr;
  
  first_buffer = new char[2 * sizeof(int) * BATCH_SIZE];
  second_buffer = new char[2 * sizeof(int) * BATCH_SIZE];

  first_fd = first_.fd;
  if(lseek(first_fd, first_.offset, SEEK_SET) < 0) return;

  second_fd = second_.fd;
  if(lseek(second_fd, second_.offset, SEEK_SET) < 0) return;

  result_fd = result_.fd;
  
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;
  
  if(read(first_fd, (char *)&first_npaths, sizeof(int)) < sizeof(int)) return;
  if(read(second_fd, (char *)&second_npaths, sizeof(int)) < sizeof(int)) return;
  result_npaths = 0;
  if(write(result_fd, (char *)&result_npaths, sizeof(int)) < sizeof(int)) return;

  read_first = true;
  read_second = true;
  i = 0; j = 0;
  int union_count=0, intersection_count=0;
  while(i < first_npaths || j < second_npaths) {
    union_count++;
    if(i < first_npaths && j < second_npaths) {
      if(read_first) {
	if(read(first_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(first_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	
	first_path = new char[first_path_len + 1];
	if(read(first_fd, (char *)first_path, first_path_len) < first_path_len) return;
	first_path[first_path_len] = 0;
	read_first = false;

      }
      if(read_second) {
	if(read(second_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(second_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	
	second_path = new char[second_path_len + 1];
	if(read(second_fd, (char *)second_path, second_path_len) < second_path_len) return;
	second_path[second_path_len] = 0;
	read_second = false;
      }

      if((condition = strcmp(first_path, second_path)) < 0) {
	num_tuples = first_ntuples;
	while(num_tuples > BATCH_SIZE) {
	  if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  num_tuples -= BATCH_SIZE;
	}

	if(num_tuples > 0) {
	  if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	}

	read_first = true;
	delete []first_path;
	i++;
      }
      else if(condition > 0) {
	num_tuples = second_ntuples;
	while(num_tuples > BATCH_SIZE) {
	  if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  num_tuples -= BATCH_SIZE;
	}
	
	if(num_tuples > 0) {
	  if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	}

	read_second = true;
	delete []second_path;
	j++;
      }
      else {
	intersection_count++;
	result_npaths++;

	if(write(result_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	num_tuples_offset = lseek(result_fd, 0, SEEK_CUR);
	if(write(result_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, second_path, second_path_len) < second_path_len) return;
		
	first_ptr.fd = first_fd;
	first_ptr.offset = lseek(first_fd, 0, SEEK_CUR);
	first_ptr.ntuples = first_ntuples;

	second_ptr.fd = second_fd;
	second_ptr.offset = lseek(second_fd, 0, SEEK_CUR);
	second_ptr.ntuples = second_ntuples;

	result_ptr.fd = result_fd;
	result_ptr.offset = num_tuples_offset + sizeof(int) + second_path_len;
	result_ptr.ntuples = 0;

	merge_adj_lists(first_ptr, second_ptr, result_ptr, distance_);

	record_end_offset = lseek(result_fd, 0, SEEK_CUR);

	if(lseek(result_fd, num_tuples_offset, SEEK_SET) < 0) return;
	result_ntuples = (record_end_offset - result_ptr.offset)/ (2 * sizeof(int));
	if(result_ntuples > 0) {
	  if(write(result_fd, (char *)&result_ntuples, sizeof(int)) < sizeof(int)) return;
	  if(lseek(result_fd, record_end_offset, SEEK_SET) < 0) return;	
	}
	else {
	  if(lseek(result_fd, num_tuples_offset-sizeof(int), SEEK_SET) < 0) return;
	  result_npaths--;
	}
	read_first = true;
	read_second = true;
	delete []first_path;
	delete []second_path;	
	i++;
	j++;
      }
    }
    else if(i < first_npaths) {
      if(read_first) {
	if(read(first_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(first_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	
	first_path = new char[first_path_len + 1];
	if(read(first_fd, (char *)first_path, first_path_len) < first_path_len) return;
	first_path[first_path_len] = 0;
      }      
      
      num_tuples = first_ntuples;
      while(num_tuples > BATCH_SIZE) {
	if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	num_tuples -= BATCH_SIZE;
      }

      if(num_tuples > 0) {
	if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
      }

      read_first = true;
      delete []first_path;
      i++;
    }
    else if(j < second_npaths) {
      if(read_second) {
	if(read(second_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(second_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	
	second_path = new char[second_path_len + 1];
	if(read(second_fd, (char *)second_path, second_path_len) < second_path_len) return;
	second_path[second_path_len] = 0;
      }

      num_tuples = second_ntuples;
      while(num_tuples > BATCH_SIZE) {
	if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	num_tuples -= BATCH_SIZE;
      }
      if(num_tuples > 0) {
	if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
      }

      read_second = true;
      delete []second_path;
      j++;
    }
  }

  if(experiment == 2) {
    StatisticsFileStream << "Union Of Paths Count In Adj Merge=" << union_count << endl;
    StatisticsFileStream << "Intersection Of Paths Count In Adj Merge=" << intersection_count << endl;
  }
  //cout << "result npath=" << result_npaths << endl;
  end_offset = lseek(result_fd, 0, SEEK_CUR);
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;
  if(write(result_fd, (char *)&result_npaths, sizeof(int)) < sizeof(int)) return;
  result_.offset = end_offset;
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;

  first_.offset = lseek(first_fd, 0, SEEK_CUR);
  second_.offset = lseek(second_fd, 0, SEEK_CUR);

  delete []first_buffer;
  delete []second_buffer;
}


void merge_lists(FileDescPtr first_ptr_, FileDescPtr second_ptr_, FileDescPtr result_ptr_) {
  int min_doc_id, doc_id, skip_count;
  int done_slot_count, i, old_rank, new_rank, same_doc_count;
  int out_array_len = first_ptr_.ntuples + second_ptr_.ntuples; // MAX_DOCS_PER_OUTPAGE; 
  doc_word_pos_struct* out_array = new (doc_word_pos_struct)[out_array_len];
  int num_lists = 2, out_docs_count;
  merge_struct* table = new (merge_struct) [num_lists];
  int *done_slots = new (int) [num_lists];
  int *same_doc_slots = new (int) [num_lists];

  lseek(result_ptr_.fd, result_ptr_.offset, SEEK_SET);

  done_slot_count = 0;
  done_slots[0] = go_first(table, 0, first_ptr_.fd, first_ptr_.offset, first_ptr_.ntuples);
  if(done_slots[0] == 1) done_slot_count++;
  done_slots[1] = go_first(table, 1, second_ptr_.fd, second_ptr_.offset, second_ptr_.ntuples);
  if(done_slots[1] == 1) done_slot_count++;  

  out_docs_count = 0;
  while(done_slot_count < num_lists) {
    same_doc_count = 0;
    skip_count = 0;
    for(i=0; i<num_lists; i++) {
      if(done_slots[i] == 1) {
	skip_count++;
	continue;
      }

      get_curr(table, i, doc_id, old_rank);
      if(skip_count == i || doc_id < min_doc_id) {
	min_doc_id = doc_id;
	new_rank = old_rank;
	same_doc_slots[0] = i;
	same_doc_count = 1;	
      }
      else if(doc_id == min_doc_id) {
	new_rank = old_rank | new_rank;	
	same_doc_slots[same_doc_count] = i;
	same_doc_count++;
      }
    }
    
    
    for(i=0; i<same_doc_count; i++) {
      if(go_next(table, same_doc_slots[i])) {
	if(done_slots[same_doc_slots[i]] == 0) {
	  done_slots[same_doc_slots[i]] = 1;
	  done_slot_count++;
	}
      }
    }

    // output min_doc_id && min_word_pos
    out_array[out_docs_count%out_array_len].doc_id = min_doc_id;
    out_array[out_docs_count%out_array_len].word_pos  = new_rank;
    out_docs_count++;
    if(out_docs_count%out_array_len == 0) {
      // print the out array
      print_out_array(out_array_len, out_array, result_ptr_.fd);
    }
  }

  if(out_docs_count%out_array_len > 0) {
    // print the out array
    print_out_array(out_docs_count%out_array_len, out_array, result_ptr_.fd);
  }

  for(i=0; i<num_lists; i++) 
    delete [](table[i].array);
  delete []table;
  delete []out_array;
  delete []done_slots;
  delete []same_doc_slots;
  return;
}

void merge_inv_lists(File_Ptr &first_, File_Ptr &second_, File_Ptr &result_) {
  int first_fd, second_fd, result_fd, i, j, condition;
  int first_npaths, second_npaths, result_npaths, end_offset, first_path_len, second_path_len;
  int first_ntuples, second_ntuples, num_tuples, result_ntuples, record_end_offset, num_tuples_offset;  
  bool read_first, read_second;
  char *first_buffer, *second_buffer, *first_path, *second_path;
  FileDescPtr first_ptr, second_ptr, result_ptr;
  
  first_buffer = new char[2 * sizeof(int) * BATCH_SIZE];
  second_buffer = new char[2 * sizeof(int) * BATCH_SIZE];

  first_fd = first_.fd;
  if(lseek(first_fd, first_.offset, SEEK_SET) < 0) return;

  second_fd = second_.fd;
  if(lseek(second_fd, second_.offset, SEEK_SET) < 0) return;

  result_fd = result_.fd;
  
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;
  
  if(read(first_fd, (char *)&first_npaths, sizeof(int)) < sizeof(int)) return;
  if(read(second_fd, (char *)&second_npaths, sizeof(int)) < sizeof(int)) return;
  result_npaths = 0;
  if(write(result_fd, (char *)&result_npaths, sizeof(int)) < sizeof(int)) return;

  read_first = true;
  read_second = true;
  i = 0; j = 0;
  int union_count = 0, intersection_count = 0;
  
  while(i < first_npaths || j < second_npaths) {
    result_npaths++;
    union_count++;
    if(i < first_npaths && j < second_npaths) {
      if(read_first) {
	if(read(first_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(first_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	
	first_path = new char[first_path_len + 1];
	if(read(first_fd, (char *)first_path, first_path_len) < first_path_len) return;
	first_path[first_path_len] = 0;
	read_first = false;

      }
      if(read_second) {
	if(read(second_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(second_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	
	second_path = new char[second_path_len + 1];
	if(read(second_fd, (char *)second_path, second_path_len) < second_path_len) return;
	second_path[second_path_len] = 0;
	read_second = false;
      }

      if((condition = strcmp(first_path, second_path)) < 0) {
	if(write(result_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, first_path, first_path_len) < first_path_len) return;

	num_tuples = first_ntuples;
	while(num_tuples > BATCH_SIZE) {
	  if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  num_tuples -= BATCH_SIZE;
	}

	if(num_tuples > 0) {
	  if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	  if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	}

	read_first = true;
	delete []first_path;
	i++;
      }
      else if(condition > 0) {
	if(write(result_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, second_path, second_path_len) < second_path_len) return;

	num_tuples = second_ntuples;
	while(num_tuples > BATCH_SIZE) {
	  if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  num_tuples -= BATCH_SIZE;
	}
	
	if(num_tuples > 0) {
	  if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	  if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	}

	read_second = true;
	delete []second_path;
	j++;
      }
      else {
	intersection_count++;
	if(write(result_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	num_tuples_offset = lseek(result_fd, 0, SEEK_CUR);
	if(write(result_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, second_path, second_path_len) < second_path_len) return;
		
	first_ptr.fd = first_fd;
	first_ptr.offset = lseek(first_fd, 0, SEEK_CUR);
	first_ptr.ntuples = first_ntuples;

	second_ptr.fd = second_fd;
	second_ptr.offset = lseek(second_fd, 0, SEEK_CUR);
	second_ptr.ntuples = second_ntuples;

	result_ptr.fd = result_fd;
	result_ptr.offset = num_tuples_offset + sizeof(int) + second_path_len;
	result_ptr.ntuples = 0;

	merge_lists(first_ptr, second_ptr, result_ptr);

	record_end_offset = lseek(result_fd, 0, SEEK_CUR);

	if(lseek(result_fd, num_tuples_offset, SEEK_SET) < 0) return;
	result_ntuples = (record_end_offset - result_ptr.offset)/ (2 * sizeof(int));
	if(write(result_fd, (char *)&result_ntuples, sizeof(int)) < sizeof(int)) return;
	if(lseek(result_fd, record_end_offset, SEEK_SET) < 0) return;	
	
	read_first = true;
	read_second = true;
	delete []first_path;
	delete []second_path;	
	i++;
	j++;
      }
    }
    else if(i < first_npaths) {
      if(read_first) {
	if(read(first_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(first_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	
	first_path = new char[first_path_len + 1];
	if(read(first_fd, (char *)first_path, first_path_len) < first_path_len) return;
	first_path[first_path_len] = 0;
      }      
      
      if(write(result_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, first_path, first_path_len) < first_path_len) return;

      num_tuples = first_ntuples;
      while(num_tuples > BATCH_SIZE) {
	if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	num_tuples -= BATCH_SIZE;
      }

      if(num_tuples > 0) {
	if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
      }

      read_first = true;
      delete []first_path;
      i++;
    }
    else if(j < second_npaths) {
      if(read_second) {
	if(read(second_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(second_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	
	second_path = new char[second_path_len + 1];
	if(read(second_fd, (char *)second_path, second_path_len) < second_path_len) return;
	second_path[second_path_len] = 0;
      }

      if(write(result_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, second_path, second_path_len) < second_path_len) return;

      num_tuples = second_ntuples;
      while(num_tuples > BATCH_SIZE) {
	if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	num_tuples -= BATCH_SIZE;
      }
      if(num_tuples > 0) {
	if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
      }

      read_second = true;
      delete []second_path;
      j++;
    }
  }

  if(experiment == 1)
    StatisticsFileStream << union_count << " "; // 1. Union Of Paths Count In Or Merge
  if(experiment == 1)
    StatisticsFileStream << intersection_count << " "; // 2. Intersection Of Paths Count In Or Merge
  
  end_offset = lseek(result_fd, 0, SEEK_CUR);
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;
  if(write(result_fd, (char *)&result_npaths, sizeof(int)) < sizeof(int)) return;
  result_.offset = end_offset;
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;

  first_.offset = lseek(first_fd, 0, SEEK_CUR);
  second_.offset = lseek(second_fd, 0, SEEK_CUR);

  delete []first_buffer;
  delete []second_buffer;
}


int find(char *in_file, char *out_file, char *keys_file, suffix_tree_node* &root, 
		   char *client_id) {
  char *path, *out_pos_file_temp, *key, *out_pos_file;
  char *out_rnk_file, *out_rnk_file_temp, *out_rnk_file_final, *copy_buffer;
  bool found, first_token = true, first_phrase = true;
  int in_fd = open(in_file, O_RDONLY), out_fd, num_documents;
  int docid, wordpos, i, count, path_len, npaths, file_offset;
  int offset, distance = 1, rank = 1, doc_pos_file, doc_rnk_file, copy_count;
  int eof_offset, last_doc_id, doc_count, doc_count_pos, token_count;    
  File_Ptr first, second, result;
  int key_count = 0;

  double insert_time = 0, insert_time_path, lseek_time, lseek_time_total = 0, dict_lookup_time, dict_lookup_time_total = 0;
  double copy_to_out_pos_time, merge_with_out_pos_time, create_out_rnk_file_time, merge_out_rnk_file_time;
  double phrase_list_gen_time, phrase_list_gen_time_total = 0, rank_list_gen_time = 0, or_time = 0;

  char phrase[MAX_PHRASE_LEN];

  struct timeval startTime;
  struct timeval stopTime;

  ifstream in(keys_file);

  out_rnk_file = new char[strlen(out_file) + strlen(".rnk") + 1];
  sprintf(out_rnk_file, "%s.rnk", out_file);

  out_rnk_file_temp = new char[strlen(out_file) + strlen(".rnk.temp") + 1];
  sprintf(out_rnk_file_temp, "%s.rnk.temp", out_file);

  out_rnk_file_final = new char[strlen(out_file) + strlen(".rnk.final") + 1];
  sprintf(out_rnk_file_final, "%s.rnk.final", out_file);

  out_pos_file = new char[strlen(out_file) + strlen(".pos") + 1];
  sprintf(out_pos_file, "%s.pos", out_file);

  out_pos_file_temp = new char[strlen(out_file) + strlen(".pos.temp") + 1];
  sprintf(out_pos_file_temp, "%s.pos.temp", out_file);  
  
  int phrase_count = 0;
  while(!in.eof()) {
    distance = 1;
    in.getline(phrase, MAX_PHRASE_LEN, '\n'); 
    if(in.gcount() == 0) break;

    //StatisticsFileStream << "Phrase=" << phrase << endl;
    phrase_count++;

    first_token = true;
    token_count = 0;
    key = strtok(phrase, " ");
    while(key) {
      token_count++;
      //StatisticsFileStream << "Key=" << key << endl;
      found = false;
      find_inv_lists(key, in_file, offset, found, dict_lookup_time); 
      dict_lookup_time_total += dict_lookup_time;
      if(!found) continue; 
      
      file_offset = offset + sizeof(RecHeader) + sizeof(int) + strlen(key) + sizeof(int); 
      
      assert(gettimeofday(&startTime, NULL) != -1);
      if(lseek(in_fd, file_offset, SEEK_SET) < 0) return 0; 
      assert(gettimeofday(&stopTime, NULL) != -1);
      lseek_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		    (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      lseek_time_total += lseek_time;
      //StatisticsFileStream << "Lseek To IF Offset Time=" << lseek_time << endl;

      if(first_token) { 
	assert(gettimeofday(&startTime, NULL) != -1);
	out_fd = open(out_pos_file, O_CREAT | O_TRUNC | O_WRONLY, S_IWRITE | S_IREAD); 
	if(read(in_fd, (char *)&npaths, sizeof(int)) < sizeof(int)) return 0; 
	if(write(out_fd, (char *)&npaths, sizeof(int)) < sizeof(int)) return 0; 
	for(int j=0; j<npaths; j++) { 
	  if(read(in_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) return 0; 
	  if(write(out_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) return 0; 
	  
	  if(read(in_fd, (char *)&count, sizeof(int)) < sizeof(int)) return 0; 
	  if(write(out_fd, (char *)&count, sizeof(int)) < sizeof(int)) return 0; 
	  
	  path = new char[path_len + 1]; 
	  if(read(in_fd, (char *)path, path_len) < path_len) return 0; 
	  if(write(out_fd, (char *)path, path_len) < path_len) return 0; 
	  delete []path; 
	  
	  copy_count = count;
	  copy_buffer = new char[BATCH_SIZE * 2 * sizeof(int)];
	  while(copy_count > BATCH_SIZE) {
	    if(read(in_fd, (char *)copy_buffer, BATCH_SIZE * 2 * sizeof(int)) < (2 * sizeof(int) * BATCH_SIZE)) return 0;
	    if(write(out_fd, (char *)copy_buffer, BATCH_SIZE * 2 * sizeof(int)) < (2 * sizeof(int) * BATCH_SIZE)) return 0;
	    copy_count -= BATCH_SIZE;
	  }
	  
	  if(copy_count > 0) {
	    if(read(in_fd, (char *)copy_buffer, copy_count * 2 * sizeof(int)) < (2 * sizeof(int) * copy_count)) return 0;
	    if(write(out_fd, (char *)copy_buffer, copy_count * 2 * sizeof(int)) < (2 * sizeof(int) * copy_count)) return 0;
	  }
	  delete []copy_buffer;
	} 
	close(out_fd); 
	assert(gettimeofday(&stopTime, NULL) != -1);
	copy_to_out_pos_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		      (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
	phrase_list_gen_time = copy_to_out_pos_time;

	//StatisticsFileStream << "Copy To Out Pos File Time=" << copy_to_out_pos_time << endl;
	first_token = false; 
      } 
      else { 
	File_Ptr second;
	second.fd = in_fd;
	second.offset = file_offset;
	
	File_Ptr first;
	first.fd = open(out_pos_file, O_RDONLY);
	first.offset = 0;
	
	File_Ptr result;
	result.fd = open(out_pos_file_temp, O_CREAT | O_TRUNC | O_WRONLY, S_IWRITE | S_IREAD);
	result.offset = 0;

	assert(gettimeofday(&startTime, NULL) != -1);
	merge_adj_inv_lists(first, second, result, distance);
	assert(gettimeofday(&stopTime, NULL) != -1);
	merge_with_out_pos_time = ((stopTime.tv_sec-startTime.tv_sec) + 
				   (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
	phrase_list_gen_time += merge_with_out_pos_time;

	//StatisticsFileStream << "Merge With Out Pos File Time=" << merge_with_out_pos_time << endl;

	close(result.fd);
	close(first.fd);

	rename(out_pos_file_temp, out_pos_file);
	distance++;
      }
      //StatisticsFileStream << endl;
      key = strtok(NULL, " ");
    }
    
    key_count += token_count;
    // StatisticsFileStream << "Token Count=" << token_count << endl;
    // StatisticsFileStream << "Phrase List Genern Time=" << phrase_list_gen_time << endl;
    phrase_list_gen_time_total += phrase_list_gen_time;

    // generate doc-rank file
    assert(gettimeofday(&startTime, NULL) != -1);
    doc_pos_file = open(out_pos_file, O_RDONLY);
    doc_rnk_file = open(out_rnk_file, O_CREAT | O_TRUNC | O_WRONLY, S_IREAD | S_IWRITE);    
    if(write(doc_rnk_file, (char *)&phrase_count, sizeof(int)) < sizeof(int)) return 0;
    if(read(doc_pos_file, (char *)&npaths, sizeof(int)) < sizeof(int)) return 0;
    if(write(doc_rnk_file, (char *)&npaths, sizeof(int)) < sizeof(int)) return 0;
    int num_path_doc_pos_triples = 0;
    char read_buf[2*sizeof(int)];
    for(int j=0; j<npaths; j++) {
      if(read(doc_pos_file, (char *)read_buf, 2*sizeof(int)) < (2*sizeof(int))) return 0;
      memcpy((char *)&path_len, (char *)read_buf, sizeof(int));
      memcpy((char *)&count, (char *)(read_buf+sizeof(int)), sizeof(int));
      
      //if(read(doc_pos_file, (char *)&path_len, sizeof(int)) < sizeof(int)) return 0;
      //if(read(doc_pos_file, (char *)&count, sizeof(int)) < sizeof(int)) return 0;
      num_path_doc_pos_triples += count;

      //if(write(doc_rnk_file, (char *)&path_len, sizeof(int)) < sizeof(int)) return 0;
      doc_count = 0;
      //doc_count_pos = lseek(doc_rnk_file, 0, SEEK_CUR);
      //if(write(doc_rnk_file, (char *)&doc_count, sizeof(int)) < sizeof(int)) return 0;      
      
      char *read_array = new char[path_len + 2 * sizeof(int) * count];
      char *write_array = new char[2*sizeof(int)+path_len + 2 * sizeof(int) * count];
      memcpy(write_array, (char *)&path_len, sizeof(int));
      if(read(doc_pos_file, (char *)read_array, path_len + 2 * sizeof(int) * count) 
	 < (path_len + 2 * sizeof(int) * count)) return 0;
      memcpy(write_array+2*sizeof(int), read_array, path_len);
      //path = new char[path_len + 1];
      //if(read(doc_pos_file, (char *)path, path_len) < path_len) return 0;
      //path[path_len] = 0;
      //if(write(doc_rnk_file, (char *)path, path_len) < path_len) return 0;
      //delete []path;
      
      //char *read_array = new char[2 * sizeof(int) * count];
      //char *write_array = new char[2 * sizeof(int) * count];
      //if(read(doc_pos_file, (char *)read_array, 2 * sizeof(int) * count) < (2 * sizeof(int) * count)) return 0;
      i = 0;
      last_doc_id = -1;
      while(i < count) {
	memcpy((char *)&docid, read_array+path_len+i*2*sizeof(int), sizeof(int));
	memcpy((char *)&wordpos, read_array+path_len+i*2*sizeof(int)+sizeof(int), sizeof(int));
	i++;
	if(docid != last_doc_id) {	  
	  last_doc_id = docid;
	  memcpy(write_array+2*sizeof(int)+path_len+doc_count*2*sizeof(int), (char *)&docid, sizeof(int));
	  memcpy(write_array+2*sizeof(int)+path_len+doc_count*2*sizeof(int)+sizeof(int), (char *)&wordpos, sizeof(int));
	  doc_count++;
	  DocsStatisticsFileStream << docid << endl;
	}
      }
      delete []read_array;

      //i = 0;
      //last_doc_id = -1;
      //while(i < count) {
      //if(read(doc_pos_file, (char *)&docid, sizeof(int)) < sizeof(int)) break;
      //if(read(doc_pos_file, (char *)&wordpos, sizeof(int)) < sizeof(int)) return 0;
      //i++;
      //if(docid != last_doc_id) {
      //last_doc_id = docid;
      //if(write(doc_rnk_file, (char *)&last_doc_id, sizeof(int)) < sizeof(int)) return 0;
      //if(write(doc_rnk_file, (char *)&rank, sizeof(int)) < sizeof(int)) return 0;
      //doc_count++;
      //}
      //}

      //eof_offset = lseek(doc_rnk_file, 0, SEEK_CUR);
      //if(lseek(doc_rnk_file, doc_count_pos, SEEK_SET) < 0) return 0;

      memcpy(write_array+sizeof(int), (char *)&doc_count, sizeof(int));

      if(write(doc_rnk_file, (char *)write_array, 2*sizeof(int)+path_len+2*sizeof(int)*doc_count) 
	 < (2*sizeof(int)+path_len+2*sizeof(int)*doc_count)) return 0;
      delete []write_array;

      //if(write(doc_rnk_file, (char *)&doc_count, sizeof(int)) < sizeof(int)) return 0;
      //if(lseek(doc_rnk_file, eof_offset, SEEK_SET) < 0) return 0;
    }

    close(doc_pos_file);
    close(doc_rnk_file);
    unlink(out_pos_file);
    assert(gettimeofday(&stopTime, NULL) != -1);
    create_out_rnk_file_time = ((stopTime.tv_sec-startTime.tv_sec) + 
				(stopTime.tv_usec-startTime.tv_usec) * 1e-6);

    // fast break in code: VLDB deadline
    // comment the next line
    StatisticsFileStream << "Num Path Doc Pos Triples=" << num_path_doc_pos_triples << endl;
    StatisticsFileStream << "Create Out Rnk File Time=" << create_out_rnk_file_time << endl;
    rank_list_gen_time += create_out_rnk_file_time;

    if(first_phrase) { 
      rename(out_rnk_file, out_rnk_file_final);
      first_phrase = false; 
    }
    else {
      // merge with existing doc_rnk_file
      first.fd = open(out_rnk_file_final, O_RDONLY);
      first.offset = sizeof(int); // after phrase_count
      
      second.fd = open(out_rnk_file, O_RDONLY);
      second.offset = sizeof(int); // after phrase_count
      
      result.fd = open(out_rnk_file_temp, O_CREAT | O_TRUNC | O_WRONLY, S_IREAD | S_IWRITE);
      if(write(result.fd, (char *)&phrase_count, sizeof(int)) < sizeof(int)) return 0;
      result.offset = sizeof(int); // after phrase_count

      assert(gettimeofday(&startTime, NULL) != -1);
      merge_inv_lists(first, second, result);
      assert(gettimeofday(&stopTime, NULL) != -1);
      merge_out_rnk_file_time = ((stopTime.tv_sec-startTime.tv_sec) + 
				  (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
      or_time += merge_out_rnk_file_time;

      if(experiment == 1)
	StatisticsFileStream << merge_out_rnk_file_time << endl; //3. Merge Out Rnk File Time 
      StatisticsFileStream << "Merge Out Rnk File Time = " << merge_out_rnk_file_time << endl; 

      close(result.fd);
      close(second.fd);
      close(first.fd);
      
      rename(out_rnk_file_temp, out_rnk_file_final);
      unlink(out_rnk_file);
    }
    //StatisticsFileStream << endl;
    rank *= 2;
  }
  close(in_fd);

  char *path_file = new char[strlen(tmp_dir) + strlen("/path_file") + strlen(client_id) + 10];
  sprintf(path_file, "%s/path_file%s.1", tmp_dir, client_id);
  int path_fd = open(path_file, O_CREAT | O_WRONLY | O_TRUNC, S_IREAD | S_IWRITE);

  doc_rnk_file = open(out_rnk_file_final, O_RDONLY);

  num_documents = 0;
  insert_time = 0;

  if(read(doc_rnk_file, (char *)&phrase_count, sizeof(int)) < sizeof(int)) return 0;
  file_offset = sizeof(int);

  if(read(doc_rnk_file, (char *)&npaths, sizeof(int)) < sizeof(int)) return 0;
  //StatisticsFileStream << "Num Paths=" << npaths << endl;
  file_offset += sizeof(int);

  for(int j=0; j<npaths; j++) {
    if(read(doc_rnk_file, (char *)&path_len, sizeof(int)) < sizeof(int)) return 0;
    if(read(doc_rnk_file, (char *)&doc_count, sizeof(int)) < sizeof(int)) return 0;

    path = new char[path_len + 1];
    if(read(doc_rnk_file, (char *)path, path_len) < path_len) return 0;
    path[path_len] = 0;    

    //StatisticsFileStream << path << " " << doc_count << endl;
    file_offset += (2 * sizeof(int) + path_len);

    
    assert(gettimeofday(&startTime, NULL) != -1);

    insert_path(root, "", path, "", file_offset, doc_count);

    if(write(path_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) return 0;
    if(write(path_fd, (char *)path, path_len) < path_len) return 0;
    if(write(path_fd, (char *)&file_offset, sizeof(int)) < sizeof(int)) return 0;
    if(write(path_fd, (char *)&doc_count, sizeof(int)) < sizeof(int)) return 0;

    assert(gettimeofday(&stopTime, NULL) != -1);

    insert_time_path = ((stopTime.tv_sec-startTime.tv_sec) + 
			(stopTime.tv_usec-startTime.tv_usec) * 1e-6);
    //StatisticsFileStream << "Insert Time for Path=" << insert_time_path << endl;
    insert_time += insert_time_path;

    num_documents += doc_count;

    if(lseek(doc_rnk_file, (doc_count * 2 * sizeof(int)), SEEK_CUR) < 0) return 0;
    file_offset += (doc_count * 2 * sizeof(int));

    //i = 0; 
    //while(i < doc_count) { 
    //if(read(doc_rnk_file, (char *)&docid, sizeof(int)) < sizeof(int)) break;
    //if(read(doc_rnk_file, (char *)&rank, sizeof(int)) < sizeof(int)) return 0;
    //StatisticsFileStream << docid << " " << rank << endl;
    //i++;
    //} 
    delete []path;
  }
  close(path_fd);
  close(doc_rnk_file);
  char *docs_file = new char[strlen(tmp_dir) + strlen("/docs_file") + strlen(client_id) + 1];
  sprintf(docs_file, "%s/docs_file%s", tmp_dir, client_id);
  rename(out_rnk_file_final, docs_file);
  delete []docs_file;

  if(experiment == 0) {
    StatisticsFileStream << npaths << " "; // 1. Num Paths
    StatisticsFileStream << insert_time << " "; // 2. Suffix Tree Insertion Time
    StatisticsFileStream << num_documents << " "; // 3. Num Path Document Pairs
    StatisticsFileStream << dict_lookup_time_total << " "; // 4. Total Dictionary Lookup Time
    StatisticsFileStream << lseek_time_total << " "; // 5. Total lseek time
    StatisticsFileStream << key_count << " "; // 6. Key Count
    StatisticsFileStream << phrase_list_gen_time_total << " "; // 7. Time to generate phrase lists
    StatisticsFileStream << rank_list_gen_time << " "; // 8. Time to generate rank list from phrase list
    StatisticsFileStream << or_time << " "; // 9. Time to perform or
    StatisticsFileStream << phrase_count << " "; // 10. Phrase Count
  }
  delete [] out_rnk_file;
  delete [] out_rnk_file_temp;
  delete [] out_rnk_file_final;
  delete [] out_pos_file;
  delete [] out_pos_file_temp;
  delete [] path_file;
  in.close();
  return 1;

}


void parse_keys(char *key_file_, char *keysinfile) {
  char *buffer;
  struct stat buf;
  int in_fd;
  bool phrase_start = false;
  ofstream out;

  out.open(keysinfile);
  if(stat(key_file_, &buf) == 0) {
    in_fd = open(key_file_, O_RDONLY);
    buffer = new char[buf.st_size + 1];
    if(read(in_fd, buffer, buf.st_size) < buf.st_size) return;
    buffer[buf.st_size] = 0;
    close(in_fd);
    for(int i=0; i<buf.st_size; i++) {
      if(phrase_start == false && buffer[i] == '"') phrase_start = true;
      else if(phrase_start == false) {
	if(buffer[i] == ' ') out << "\n";
	else out << buffer[i];
      }
      else if(phrase_start == true && buffer[i] == '"') phrase_start = false;
      else {
	out << buffer[i];
      }
    }

    delete []buffer;
  }
  else exit(0);
  out.close();
}

// args: /mjolnir5/thomas/index/index.final, /tmp/out, keysfile, client id, /mjolnir/thomas/research/find, /tmp
//      (final index file), (prefix of all out files), (input terms), (unique id for client)
int main(int argc, char* argv[]) {
  if(argc < 7) {
    exit(0);
  }

  char *client_id = argv[4];

  int log_dir_len = strlen(argv[5]);
  log_dir = new char[log_dir_len + 1];
  strcpy(log_dir, argv[5]);
  if((log_dir_len > 1) && (log_dir[log_dir_len-1] == '/')) log_dir[log_dir_len-1] = 0;
  log_dir_len--;

  int tmp_dir_len = strlen(argv[6]);
  char *tmp_dir = new char[tmp_dir_len + 1];
  strcpy(tmp_dir, argv[6]);
  if((tmp_dir_len > 1) && (tmp_dir[tmp_dir_len-1] == '/')) tmp_dir[tmp_dir_len-1] = 0;
  tmp_dir_len--;
  
  char *fname = new char[strlen(log_dir) + strlen("/find_statistics.log") + 10];
  sprintf(fname, "%s/find_statistics.log.%d", log_dir, experiment);
  StatisticsFileStream.open(fname, ios::out | ios::app);
  StatisticsFileStream.setf(ios::fixed, ios::floatfield);
  delete []fname;

  
  fname = new char[strlen(log_dir) + strlen("/docs_statistics.log") + 10];
  sprintf(fname, "%s/docs_statistics.log.%d", log_dir, rand()%10000);
  DocsStatisticsFileStream.open(fname, ios::out | ios::app);
  DocsStatisticsFileStream.setf(ios::fixed, ios::floatfield);
  delete []fname;

  suffix_tree_node *root = NULL;
  char *sub_file = new char[tmp_dir_len + strlen("/sub_file") + strlen(client_id) + 10];
  sprintf(sub_file, "%s/sub_file%s.1", tmp_dir, client_id);

  struct timeval startTime;
  struct timeval stopTime;

  char *out_file = new char[strlen(argv[2]) + strlen(client_id) + 1];
  sprintf(out_file, "%s%s", argv[2], client_id);

  char *keysinfile = new char[tmp_dir_len + strlen("/keysin") + strlen(client_id) + 1];
  sprintf(keysinfile, "%s/keysin%s", tmp_dir, client_id);
  parse_keys(argv[3], keysinfile);
  find(argv[1], out_file, keysinfile, root, client_id);
  
  struct stat buf;
  if(stat(argv[3], &buf) == 0) {
    int in_fd = open(argv[3], O_RDONLY);
    char *buffer = new char[buf.st_size + 1];
    if(read(in_fd, buffer, buf.st_size) < buf.st_size) return 0;
    buffer[buf.st_size] = 0;
    close(in_fd);
    if(experiment == 0)
      StatisticsFileStream << "#Query = " << buffer << endl;
    DocsStatisticsFileStream << "--------------#Query = " << buffer << endl;
  }
  // print tree to file
  double print_file_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);

  if(root) print_to_file(sub_file, root);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_file_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  if(experiment == -1)
    StatisticsFileStream << print_file_time << " "; // 1. Suffix Tree Print To File Time

  // print tree as html
  int init_view = 1;
  double print_html_time = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  int max_sup_level = 0, max_sub_level = 0;
  int sub_node_count = 0, sup_node_count = 0;
  // SIGMOD deadline: fast break in code
  // uncomment the next line before launching system - Thomas
  print_html_trees(init_view, 0, NULL, root, client_id, sub_node_count, max_sub_level, sup_node_count, max_sup_level);
  assert(gettimeofday(&stopTime, NULL) != -1);
  print_html_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		     (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  
  if(experiment == -1) 
    StatisticsFileStream << sub_node_count<< " "; // 2.
  if(experiment == -1) 
    StatisticsFileStream << max_sup_level<< " "; // 3. 
  if(experiment == -1) 
    StatisticsFileStream << sup_node_count<< " "; // 4. 
  if(experiment == -1) 
    StatisticsFileStream << max_sub_level<< " "; // 5.
  if(experiment == -1)
    StatisticsFileStream << print_html_time << " "; // 6. Suffix Tree HTML Print Time

  // delete tree
  double delete_tree_time;
  assert(gettimeofday(&startTime, NULL) != -1);
  delete_suffix_tree(root);
  assert(gettimeofday(&stopTime, NULL) != -1);
  delete_tree_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		      (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  if(experiment == -1)
    StatisticsFileStream << delete_tree_time << endl; // 7. delete tree time

  StatisticsFileStream.close();
  DocsStatisticsFileStream.close();
  delete []sub_file;
  delete []out_file;
  delete []keysinfile;
  delete []log_dir;
  delete []tmp_dir;
}

/*

main(int argc, char* argv[]) {
  char *in_file1 = argv[1], *in_file2 = argv[2];
  char *out_file = argv[3], *path;
  File_Ptr first, second, result;
  int in_fd, count, path_len, npaths, i, docid, wordpos;

  first.fd = open(in_file1, O_RDONLY);
  first.offset = 0;

  second.fd = open(in_file2, O_RDONLY);
  second.offset = 0;

  result.fd = open(out_file, O_CREAT | O_TRUNC | O_WRONLY, S_IREAD | S_IWRITE);
  result.offset = 0;

  merge_adj_inv_lists(first, second, result, 1);

  close(first.fd);
  close(second.fd);
  close(result.fd);

  in_fd = open(out_file, O_RDONLY);
  
  if(read(in_fd, (char *)&npaths, sizeof(int)) < sizeof(int)) return 0;
  cout << npaths << endl;
  
  for(int j=0; j<npaths; j++) {
    if(read(in_fd, (char *)&path_len, sizeof(int)) < sizeof(int)) return 0;
    if(read(in_fd, (char *)&count, sizeof(int)) < sizeof(int)) return 0;
    
    path = new char[path_len + 1];
    if(read(in_fd, (char *)path, path_len) < path_len) return 0;
    path[path_len] = 0;
    
    cout << path << " " << count << endl;
    delete []path;
    
    i = 0;
    while(i < count) {
      if(read(in_fd, (char *)&docid, sizeof(int)) < sizeof(int)) break;
      if(read(in_fd, (char *)&wordpos, sizeof(int)) < sizeof(int)) return 0;
      cout << docid << " " << wordpos << endl;
      i++;
    }
  }
 
  close(in_fd);
  return 1;
}
  

*/
