
#include "InvList.h"
#define ESIZE 100
#include "regex.h"
#include <fstream.h>
#include <sys/time.h> //gettimeofday, struct timeval

extern ofstream GLOBAL_logFileStream;
bool GLOBAL_log = false;

//This block of code is for regular expression support
#define INIT         register char *sp=instring;
#define GETC()       (*sp++)
#define PEEKC()      (*sp)
#define UNGETC(c)     (--sp)
#define RETURN(c)    return 0;
#define ERROR(c)     exit(0);
#include <regexp.h>

char *compile(char *instring, char *expbuf,
	      const char *endbuf, int eof);

int step(const char *string, const char *expbuf);

int advance(const char *string, const char *expbuf);

extern char *loc1, *loc2, *locs;
//////////////////////////////////////////////////////

InvList::InvList(char *buf_ = 0, int len_ = 0) {
  _first = 0;
  _curr_path_entry = NULL;
  _curr_doc_entry = NULL;
  _npaths   = 0;
  _print_len = sizeof(int);
  _num_entries = 0;

  if(len_ > 0) {
    char *ptr = buf_;
    int npaths, path_len, ntuples;
    int doc_id, word_posn;
    char *path;
    memcpy((char *)&npaths, ptr, sizeof(int));
    ptr += sizeof(int);

    for(int i=0; i<npaths; i++) {
      memcpy((char *)&path_len, ptr, sizeof(int));
      ptr += sizeof(int);

      memcpy((char *)&ntuples, ptr, sizeof(int));
      ptr += sizeof(int);

      path = new char[path_len+1];
      memcpy(path, ptr, path_len);
      path[path_len] = 0;
      ptr += path_len;
      
      for(int ii=0; ii<ntuples; ii++) {
	memcpy((char *)&doc_id, ptr, sizeof(int));
	ptr += sizeof(int);

	memcpy((char *)&word_posn, ptr, sizeof(int));
	ptr += sizeof(int);

	append(path, doc_id, word_posn);
      }
      
      delete []path;
    }
  }
}

InvList::InvList(const InvList& other_) {
  _first = 0;
  _curr_path_entry = NULL;
  _curr_doc_entry = NULL;
  _npaths = 0;
  _num_entries = 0;
  _print_len = sizeof(int);
  *this = other_;
}

InvList::~InvList() {
  free();
}

void InvList::free() {
  PEntry *path_entry = _first;
  while(path_entry) {
    PEntry *prev_path_entry = path_entry;
    path_entry = path_entry->pNext;

    DEntry *doc_entry = prev_path_entry->dFirst;
    while(doc_entry) {
      DEntry *prev_doc_entry = doc_entry;
      doc_entry = doc_entry->dNext;
      delete prev_doc_entry;
    }
    
    delete [](prev_path_entry->pname);
    delete prev_path_entry;
  }
  _first = 0;
  _curr_path_entry = NULL;
  _curr_doc_entry = NULL;
  _npaths = 0;
  _print_len = sizeof(int);
  _num_entries = 0;
}

InvList& InvList::operator =(const InvList& other_) {
  struct timeval startTime;
  struct timeval stopTime;

  assert(gettimeofday(&startTime, NULL) != -1);
  if(this != &other_) {
    this->free();
    PEntry *path_entry = other_._first;
    while(path_entry) {
      DEntry *doc_entry = path_entry->dFirst;
      while(doc_entry) {
	this->append(path_entry->pname, doc_entry->doc_id, doc_entry->word_pos);
	doc_entry = doc_entry->dNext;
      }

      path_entry = path_entry->pNext;            
    }
  }
  assert(gettimeofday(&stopTime, NULL) != -1);
  double copyTime = 
    (stopTime.tv_sec-startTime.tv_sec) + 
    (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
  if(GLOBAL_log) GLOBAL_logFileStream << "IL copy time: " << copyTime << " seconds" << endl;
  return *this;
}


void InvList::append(char* pname_, int doc_id_, int word_pos_) {
  if(_curr_path_entry && strcmp(_curr_path_entry->pname, pname_) == 0) {
    DEntry *temp = _curr_doc_entry;
    
    _curr_doc_entry = new DEntry;
    _curr_doc_entry->doc_id = doc_id_;
    _curr_doc_entry->word_pos = word_pos_;
    _curr_doc_entry->dNext = NULL;
    (_curr_path_entry->count)++;
    
    if(temp) 
      temp->dNext = _curr_doc_entry;
    else
      _curr_path_entry->dFirst = _curr_doc_entry;

    _curr_path_entry->dLast = _curr_doc_entry;

    _print_len = _print_len + (2 * sizeof(int)); // docid, wordpos
    _num_entries++;
  }
  else {
    PEntry *temp = _curr_path_entry;

    _curr_path_entry = new PEntry;
    _curr_path_entry->pname = new char[strlen(pname_)+1];
    strcpy(_curr_path_entry->pname, pname_);
    _curr_path_entry->count = 1;

    _curr_doc_entry = new DEntry;
    _curr_doc_entry->doc_id = doc_id_;
    _curr_doc_entry->word_pos = word_pos_;
    _curr_doc_entry->dNext = NULL;
    _curr_path_entry->dFirst = _curr_doc_entry;
    _curr_path_entry->dLast = _curr_doc_entry;

    _curr_path_entry->pNext = NULL;

    if(temp)
      temp->pNext = _curr_path_entry;
    else
      _first = _curr_path_entry;

    _npaths++;
    _print_len = _print_len + (sizeof(int) + sizeof(int) + strlen(pname_) + 2 * sizeof(int)); 
    // ntuples, pathlen, path, docid, wordpos
    _num_entries++;
  }
}

void InvList::insert(char* pname_, int doc_id_, int word_pos_) {
  
  DEntry* new_doc_entry = new DEntry;
  new_doc_entry->doc_id = doc_id_;
  new_doc_entry->word_pos = word_pos_;
  new_doc_entry->dNext = 0;
  _print_len = _print_len + sizeof(int) * 2;
  _num_entries++;

  PEntry *path_entry = _first;
  PEntry *prev_path_entry = 0;

  while(path_entry) {
    int condition = strcmp(path_entry->pname, pname_);
    if(condition == 0) {
      path_entry->count++; 

      if(path_entry->dLast)
	path_entry->dLast->dNext = new_doc_entry;
      else 
	path_entry->dFirst = new_doc_entry;

      path_entry->dLast = new_doc_entry;

      /*
	Making use of the fact that insertions are in increasing order of docid, wordpos pairs

	DEntry *doc_entry = path_entry->dFirst;
	DEntry *prev_doc_entry = 0;
	while(doc_entry) {
	if(doc_entry->doc_id < doc_id_ || ((doc_entry->doc_id == doc_id_) && (doc_entry->word_pos < word_pos_))) {
	prev_doc_entry = doc_entry;
	doc_entry = doc_entry->dNext;
	} 
	else if(doc_entry->doc_id > doc_id_ || ((doc_entry->doc_id == doc_id_) && (doc_entry->word_pos > word_pos_))) {
	break;
	}
	else {
	GLOBAL_logFileStream << "InvList: " << "Error: This is not possible - matching doc, word pair" << endl;
	return;
	}
	}
	if(prev_doc_entry)
	prev_doc_entry->dNext = new_doc_entry;
	else
	path_entry->dFirst = new_doc_entry;
	new_doc_entry->dNext = doc_entry;
	*/
      
      return;
    }
    else if(condition < 0) {
      prev_path_entry = path_entry;
      path_entry = path_entry->pNext;
    }
    else
      break;
  }

  PEntry* new_path_entry = new PEntry;
  new_path_entry->pname = new char[strlen(pname_) + 1];
  memcpy(new_path_entry->pname, pname_, strlen(pname_) + 1);
  new_path_entry->count = 1;
  new_path_entry->dFirst = new_doc_entry;
  new_path_entry->dLast = new_doc_entry;
  if(prev_path_entry)
    prev_path_entry->pNext = new_path_entry;
  else
    _first = new_path_entry;
  new_path_entry->pNext = path_entry;
  _npaths++;
  _print_len = _print_len + (sizeof(int) + strlen(pname_) + sizeof(int)); // numtuples, path, pathlen
}

void InvList::remove(int doc_id_) {
  PEntry* path_entry = _first;
  while(path_entry) {
    DEntry* prev_doc_entry = NULL;
    DEntry* doc_entry = path_entry->dFirst;
    while(doc_entry) {
      if(doc_entry->doc_id == doc_id_) {
	_print_len -= (2 * sizeof(int));
	_num_entries--;
	if(prev_doc_entry)
	  prev_doc_entry->dNext  = doc_entry->dNext;
	else
	  path_entry->dFirst = doc_entry->dNext;

	if(doc_entry == path_entry->dLast)
	  path_entry->dLast = prev_doc_entry;

	DEntry* del_doc_entry = doc_entry;
	doc_entry = doc_entry->dNext;
	delete del_doc_entry;
	path_entry->count--; 
      }
      else {
	prev_doc_entry = doc_entry;
	doc_entry = doc_entry->dNext;
      }
    }
    path_entry = path_entry->pNext;
  }
}

int InvList::get_print_len() const {
  return _print_len;
}

int InvList::get_num_entries() const {
  return _num_entries;
}

char* InvList::toCharPtr() const {
  char *buf = new char[_print_len];
  char *ptr = buf;

  memcpy(ptr, (char *)&_npaths, sizeof(int));
  ptr += sizeof(int);

  PEntry *temp = _first;

  while(temp) {
    int pathlen = strlen(temp->pname);
    memcpy(ptr, (char *)&pathlen, sizeof(int));
    ptr += sizeof(int);

    int numtuples = temp->count;
    memcpy(ptr, (char *)&numtuples, sizeof(int));
    ptr += sizeof(int);

    memcpy(ptr, temp->pname, pathlen);
    ptr += pathlen;;

    DEntry *docPtr = temp->dFirst;
    while(docPtr) {
      memcpy(ptr, (char *)(&(docPtr->doc_id)), sizeof(int));
      ptr += sizeof(int);
      
      memcpy(ptr, (char *)(&(docPtr->word_pos)), sizeof(int));
      ptr += sizeof(int);

      docPtr = docPtr->dNext;
    } 
    temp = temp->pNext;      
  }
  return buf;
}

bool InvList::isEmpty() const {
  if(_first)
    return 0;
  return 1;
}

InvList adjacent(InvList &curr, InvList &other_, int distance_) {
  InvList result;

  PEntry *curr_path_entry  = curr._first;
  PEntry *other_path_entry = other_._first;
  
  while(curr_path_entry && other_path_entry) {
    int condition = strcmp(curr_path_entry->pname, other_path_entry->pname);
    if(condition == 0) {
      DEntry *curr_doc_entry  = curr_path_entry->dFirst;
      DEntry *other_doc_entry = other_path_entry->dFirst;
      while(curr_doc_entry && other_doc_entry) {
	if(curr_doc_entry->doc_id < other_doc_entry->doc_id)
	  curr_doc_entry = curr_doc_entry->dNext;
	else if(curr_doc_entry->doc_id > other_doc_entry->doc_id)
	  other_doc_entry = other_doc_entry->dNext;
	else {
	  if(curr_doc_entry->word_pos + distance_ < other_doc_entry->word_pos)
	    curr_doc_entry = curr_doc_entry->dNext;
	  else if(curr_doc_entry->word_pos + distance_ > other_doc_entry->word_pos)
	    other_doc_entry = other_doc_entry->dNext;
	  else
	    result.append(curr_path_entry->pname, curr_doc_entry->doc_id, curr_doc_entry->word_pos);
	}
      }
      curr_path_entry = curr_path_entry->pNext;
      other_path_entry = other_path_entry->pNext;
    }
    else if(condition < 0) {
      curr_path_entry = curr_path_entry->pNext;
    }
    else {
      other_path_entry = other_path_entry->pNext;
    }
  }
  return result;
}



InvList InvList::operator +(const InvList& other_) {
  InvList result;
  struct timeval startTime;
  struct timeval stopTime;

  assert(gettimeofday(&startTime, NULL) != -1);

  PEntry *lhs_entry = _first;
  PEntry *rhs_entry = other_._first;
  DEntry *doc_entry;

  while(lhs_entry || rhs_entry) {    
    if(lhs_entry && rhs_entry) {
      if(strcmp(lhs_entry->pname, rhs_entry->pname) < 0) {
	doc_entry = lhs_entry->dFirst;
	while(doc_entry) {
	  if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
					      << (lhs_entry->pname) << ":(" 
					      << doc_entry->doc_id << ":" 
					      << doc_entry->word_pos << ")"
					      << endl;

	  result.append(lhs_entry->pname, doc_entry->doc_id, doc_entry->word_pos);
	  doc_entry = doc_entry->dNext;
	}
	lhs_entry = lhs_entry->pNext;  
      }
      else if(strcmp(lhs_entry->pname, rhs_entry->pname) > 0) {
	doc_entry = rhs_entry->dFirst;
	while(doc_entry) {
	  if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
					      << (rhs_entry->pname) << ":(" 
					      << doc_entry->doc_id << ":" 
					      << doc_entry->word_pos << ")"
					      << endl;

	  result.append(rhs_entry->pname, doc_entry->doc_id, doc_entry->word_pos);
	  doc_entry = doc_entry->dNext;
	}
	rhs_entry = rhs_entry->pNext;
      }
      else {
	DEntry *lhs_doc_entry = lhs_entry->dFirst;
	DEntry *rhs_doc_entry = rhs_entry->dFirst;
	
	while(lhs_doc_entry || rhs_doc_entry) {    
	  if(lhs_doc_entry && rhs_doc_entry) {
	    if(lhs_doc_entry->doc_id < rhs_doc_entry->doc_id) {
	      if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
						  << (lhs_entry->pname) << ":(" 
						  << lhs_doc_entry->doc_id << ":" 
						  << lhs_doc_entry->word_pos << ")"
						  << endl;

	      result.append(lhs_entry->pname, lhs_doc_entry->doc_id, lhs_doc_entry->word_pos);
	      lhs_doc_entry = lhs_doc_entry->dNext;
	    }
	    else if(lhs_doc_entry->doc_id > rhs_doc_entry->doc_id) {
	      if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
						  << (rhs_entry->pname) << ":(" 
						  << rhs_doc_entry->doc_id << ":" 
						  << rhs_doc_entry->word_pos << ")"
						  << endl;

	      result.append(rhs_entry->pname, rhs_doc_entry->doc_id, rhs_doc_entry->word_pos);
	      rhs_doc_entry = rhs_doc_entry->dNext;	      
	    }
	    else {	  
	      if(lhs_doc_entry->word_pos < rhs_doc_entry->word_pos) {
		if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
						    << (lhs_entry->pname) << ":(" 
						    << lhs_doc_entry->doc_id << ":" 
						    << lhs_doc_entry->word_pos << ")"
						    << endl;

		result.append(lhs_entry->pname, lhs_doc_entry->doc_id, lhs_doc_entry->word_pos);
		lhs_doc_entry = lhs_doc_entry->dNext;
	      }
	      else {
		if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
						    << (rhs_entry->pname) << ":(" 
						    << rhs_doc_entry->doc_id << ":" 
						    << rhs_doc_entry->word_pos << ")"
						    << endl;

		result.append(rhs_entry->pname, rhs_doc_entry->doc_id, rhs_doc_entry->word_pos);
		rhs_doc_entry = rhs_doc_entry->dNext;
	      }
	    }
	  }
	  else if(lhs_doc_entry) {
	    while(lhs_doc_entry) {
	      if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
						  << (lhs_entry->pname) << ":(" 
						  << lhs_doc_entry->doc_id << ":" 
						  << lhs_doc_entry->word_pos << ")"
						  << endl;

	      result.append(lhs_entry->pname, lhs_doc_entry->doc_id, lhs_doc_entry->word_pos);
	      lhs_doc_entry = lhs_doc_entry->dNext;
	    }
	  }
	  else {
	    while(rhs_doc_entry) {
	      if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
						  << (rhs_entry->pname) << ":(" 
						  << rhs_doc_entry->doc_id << ":" 
						  << rhs_doc_entry->word_pos << ")"
						  << endl;
	      
	      result.append(rhs_entry->pname, rhs_doc_entry->doc_id, rhs_doc_entry->word_pos);
	      rhs_doc_entry = rhs_doc_entry->dNext;
	    }
	  }
	} // while(lhs_doc_entry || rhs_entry) 	

	lhs_entry = lhs_entry->pNext;
	rhs_entry = rhs_entry->pNext;      
      }
    }
    else if(lhs_entry) {
      while(lhs_entry) {
	doc_entry = lhs_entry->dFirst;
	while(doc_entry) {
	  if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
					      << (lhs_entry->pname) << ":(" 
					      << doc_entry->doc_id << ":" 
					      << doc_entry->word_pos << ")"
					      << endl;

	  result.append(lhs_entry->pname, doc_entry->doc_id, doc_entry->word_pos);
	  doc_entry = doc_entry->dNext;
	}
	lhs_entry = lhs_entry->pNext;
      }
    }
    else {
      while(rhs_entry) {
	doc_entry = rhs_entry->dFirst;
	while(doc_entry) {
	  if(GLOBAL_log) GLOBAL_logFileStream << "InvList: " << "appending " 
					      << (rhs_entry->pname) << ":(" 
					      << doc_entry->doc_id << ":" 
					      << doc_entry->word_pos << ")"
					      << endl;

	  result.append(rhs_entry->pname, doc_entry->doc_id, doc_entry->word_pos);
	  doc_entry = doc_entry->dNext;
	}
	rhs_entry = rhs_entry->pNext;
      }
    }
  } // while(lhs_entry || rhs_entry) 

  assert(gettimeofday(&stopTime, NULL) != -1);
  double mergeTime = 
    (stopTime.tv_sec-startTime.tv_sec) + 
    (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
  if(GLOBAL_log) GLOBAL_logFileStream << "IL merge time: " << mergeTime << " seconds" << endl;
  return result;
}

InvList InvList::prune(char *path_) const{
  InvList result;
  char expbuf[ESIZE];
  char regString[ESIZE];

  if(createRegExp(path_, regString, ESIZE) == -1)
    return result; 
  compile((char *)regString, (char *)expbuf, (const char *)(&expbuf[ESIZE]), '\0');
  
  PEntry *path_entry = _first;
  while(path_entry) {
    PEntry *prev_path_entry = path_entry;
    path_entry = path_entry->pNext;
    
    DEntry *doc_entry = prev_path_entry->dFirst;
    if(step((const char *)(prev_path_entry->pname), (const char *)expbuf)) {
      while(doc_entry) {
	DEntry *prev_doc_entry = doc_entry;
	doc_entry = doc_entry->dNext;
	result.append(prev_path_entry->pname, prev_doc_entry->doc_id, prev_doc_entry->word_pos);
      }      
    }
  }

  return result;
}


List<Int, int> InvList::condense() const{
  List<Int, int> result;
  
  PEntry *ptr = _first;
  while(ptr) {
    DEntry *docPtr = ptr->dFirst;
    while(docPtr) {
      result.insert(Int(docPtr->doc_id));
      docPtr = docPtr->dNext;
    }
    ptr = ptr->pNext;
  }
  return result;
}

void InvList::goFirst() {
  _curr_path_entry = _first;
  _curr_doc_entry = NULL;
  while(_curr_path_entry) {
    _curr_doc_entry = _curr_path_entry->dFirst;
    if(_curr_doc_entry) break;
    _curr_path_entry = _curr_path_entry->pNext;
  }
  return;
}

void InvList::goNext() {
  bool found = false;
  int prev_doc_id = _curr_doc_entry->doc_id;
  _curr_doc_entry  = _curr_doc_entry->dNext;
  while(_curr_path_entry) {
    while(_curr_doc_entry) {
      if(_curr_doc_entry->doc_id != prev_doc_id) {
	found = true; 
	break;
      }
      _curr_doc_entry = _curr_doc_entry->dNext;
    }
    if(found) break;
    _curr_path_entry = _curr_path_entry->pNext;
    if(_curr_path_entry)
      _curr_doc_entry  = _curr_path_entry->dFirst;
  }
  return;
}

bool InvList::isOffList() const {
  if(!_curr_path_entry && !_curr_doc_entry) return true;
  return false;
}

int InvList::getFromCurrent(char *& path_name_, int &doc_id_) const {
  if(_curr_path_entry && _curr_doc_entry) {
    path_name_ = new char[strlen(_curr_path_entry->pname)+1];
    strcpy(path_name_, _curr_path_entry->pname);
    doc_id_ = _curr_doc_entry->doc_id;
    return 0;
  }
  return -1;
}

