#include "InvertedFile5.h"
#include <strings.h>   // strcpy, strlen
#include <sys/types.h> // stat, lseek, open
#include <sys/stat.h>  // stat, open
#include <stdio.h>     // sprintf
#include <assert.h>    // assert
#include <unistd.h>    // lseek, read, write
#include <fcntl.h>     // open
#include <errno.h>     // ENOENT
#include <fstream>     // ofstream
#include "key_file_record.h" // KEY_SIZE

extern "C" int errno;
#define MAX_KEY_LEN (KEY_SIZE)

extern ofstream GLOBAL_logFileStream;
static bool GLOBAL_log = false;

int InvertedFile::my_write(int fd_, char *ptr_, int len_) {
  return write(fd_, ptr_, len_);
}

InvertedFile::InvertedFile(char *fname_, char *idString_ = "IF: ") {
  _idString = new char[1 + strlen(idString_)];
  strcpy(_idString, idString_);

  _if_fname = new char[strlen(fname_) + 1];
  if(NULL == _if_fname) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": malloc in constr failed" << endl;
    return;
  }
  strcpy(_if_fname, fname_);

  _if_iname = new char[strlen(_if_fname) + strlen(".idx") + 1];
  if(NULL == _if_iname) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": malloc in constr failed" << endl;
    return;
  }
  sprintf(_if_iname, "%s.idx", _if_fname);

  char *idString = new char[1 + strlen(_idString) + strlen("HT.")];
  sprintf(idString, "HT.%s", _idString);

  delete []idString;

  struct stat buf;
  if( (stat(_if_fname, &buf) < 0) && (errno == ENOENT)) {
    _fd = open(_if_fname, O_CREAT | O_RDWR, S_IRWXU);
    if(_fd < 0) {
      if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": constr; file creation failed" << endl;
      return;
    }

    _fheader.first = -1;
    _fheader.end_of_file = sizeof(FileHeader);

    if(my_write(_fd, (char *)&_fheader, sizeof(FileHeader)) < sizeof(FileHeader)) {
      if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": file header write in constr failed" << endl;
      return;
    }
    
    close(_fd);
  }

  _fd = open(_if_fname, O_RDWR);    
  if(_fd < 0) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": open file in constr failed" << endl;
    return;
  }
  if(read(_fd, (char *)&_fheader, sizeof(FileHeader)) < sizeof(FileHeader)) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": file header read in constr failed" << endl;
    return;
  }
  
  _modified = false;
}

InvertedFile::InvertedFile(const InvertedFile& other) {
  _idString = new char[1 + strlen(other._idString)];
  strcpy(_idString, other._idString);

  _if_fname = new char[strlen(other._if_fname) + 1];
  if(NULL == _if_fname) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": malloc in copy constr failed" << endl;
    return;
  }
  strcpy(_if_fname, other._if_fname);

  _if_iname = new char[strlen(_if_fname) + strlen(".idx") + 1];
  if(NULL == _if_iname) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": malloc in copy constr failed" << endl;
    return;
  }
  sprintf(_if_iname, "%s.idx", _if_fname);

  char *idString = new char[1 + strlen(_idString) + strlen("HT.")];
  sprintf(idString, "HT.%s", _idString);

  delete []idString;

  struct stat buf;
  if( (stat(_if_fname, &buf) < 0) && (errno == ENOENT)) {
    _fd = open(_if_fname, O_CREAT | O_RDWR, S_IRWXU);
    if(_fd < 0) {
      if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": file create in copy constr failed" << endl;
      return;
    }
    _fheader.first = -1;
    _fheader.end_of_file = sizeof(FileHeader);

    if(my_write(_fd, (char *)&_fheader, sizeof(FileHeader)) < sizeof(FileHeader)) {
      if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": file header write in copy constr failed" << endl;
      return;
    }
    close(_fd);
  }
  
  _fd = open(_if_fname, O_RDWR);
  if(_fd < 0) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": open file in copy constr failed" << endl;
    return;
  }
  if(read(_fd, (char *)&_fheader, sizeof(FileHeader)) < sizeof(FileHeader)) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": file header read in copy constr failed" << endl;
    return;
  }
  
  _modified = false;  
}

InvertedFile::~InvertedFile() {
  if(_modified) {
    if(lseek(_fd, 0, SEEK_SET) < 0) {
      if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": lseek in destructor failed" << endl;
      return;
    }
    if(my_write(_fd, (char *)&_fheader, sizeof(FileHeader)) < sizeof(FileHeader)) {
      if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": write of file header in destructor failed" << endl;
      return;
    }
  }
  close(_fd);
  delete []_if_fname;
  delete []_if_iname;
  delete []_idString;
}

int InvertedFile::insert_record(String key_, const InvList &list_, Time time_) {
  RecHeader rh;
  int rec_offset, nbytes;

  // find slot to insert record
  rec_offset = lseek(_fd, 0, SEEK_END);    
  if(rec_offset < 0) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": insert; lseek to end of file failed" << endl;
    return -1;
  }
  
  rh.free = 0;
  char *ptr = time_.toCharPtr();
  memcpy((char *)&(rh.sec), ptr, sizeof(long));
  memcpy((char *)&(rh.usec), ptr+sizeof(long), sizeof(long));
  delete []ptr;
  rh.next = -1;
  
  _modified = true;
  
  // insert the invlist
  int word_len = key_.get_len();
  char *word_ptr = key_.toCharPtr();

  int list_len = list_.get_print_len();
  char *list_data = list_.toCharPtr();

  int   record_len = sizeof(RecHeader) + sizeof(int) * 2 + word_len + list_len;
  char *record_ptr = new char[record_len];
  char *tmp_rcd_ptr= record_ptr;

  _fheader.end_of_file += record_len;
  
  memcpy(tmp_rcd_ptr, (char *)&rh, sizeof(RecHeader));
  tmp_rcd_ptr +=  sizeof(RecHeader);

  memcpy(tmp_rcd_ptr, (char *)&word_len, sizeof(int));
  tmp_rcd_ptr += sizeof(int);

  memcpy(tmp_rcd_ptr, word_ptr, word_len);
  tmp_rcd_ptr += word_len;

  memcpy(tmp_rcd_ptr, (char *)&list_len, sizeof(int));
  tmp_rcd_ptr += sizeof(int);

  memcpy(tmp_rcd_ptr, list_data, list_len);
  tmp_rcd_ptr += list_len;

  nbytes = my_write(_fd, record_ptr, record_len);
  if(nbytes < record_len) {
    if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": insert; write of inverted list failed. len = " << list_len << endl;
    delete []list_data;
    delete []record_ptr;
    return -1;
  }
  delete []word_ptr;
  delete []list_data;
  delete []record_ptr;

  return 0;
}

