			/* ==== :INFO/STRIP.C ==== */

/*
 * AUTHOR	: Stuart J. Barr 
 * DATE:	: c. September 1986, give or take a few months...
 */

/*
 * to save confusion...
 * USAGE	: define KEYWORDSIZE (6 is a good value, in practise)
 *		  define FALSE, TRUE
 *		  if you want to strip prefixes, define PREFIXES
 *		  Write a wee C function to call
 *			strip_affixes(word)
 *		  where word is a char *.
 */

#include "porter.h"

void strip_affixes (char* word )
{
	to_lower_case(word);
	clean(word);
#ifdef PREFIXES
	/*strip_prefixes(word);*/
#endif PREFIXES
	/*strip_suffixes(word);*/
	/*word[KEYWORDSIZE] = '\0';*/
}

 void to_lower_case (char* kwd )
{
	int i;
	for ( i=0 ; i < strlen(kwd) ; i++ )
		if ( (kwd[i] >= 'A') && (kwd[i] <= 'Z') )
			kwd[i] += 'a' - 'A';
}

 void clean (char* kwd )
{
	int i,j,last=strlen(kwd);
	for ( i=0 ; i < last ; i++ ) {
		if ( isvalid(kwd[i]) != 0 ) {
			for ( j = i ; j < last-1 ; j++ )
				kwd[j] = kwd[j+1];
			kwd[last-1] = '\0';
			last--;
			i--;
		}
	}
}

 isvalid(char l)
{
	if ( (l >= 'a') && (l <= 'z') )
		return(0);
	if ( (l >= 'A') && (l <= 'Z') )
		return(0);
	if ( (l >= '0') && (l <= '9') )
		return(0);
	return(1);
}

#ifdef PREFIXES
 void strip_prefixes (char* word )
{
	 char *prefixes[] = { "kilo", "micro", "milli", "intra", "ultra",
					"mega", "nano", "pico", "pseudo", 0 };
	int i;
	for ( i=0 ; prefixes[i] != 0 ; i++ ) {
		if ( strncmp(prefixes[i],word,strlen(prefixes[i])) == 0 ) {
			int j;
			for ( j=0 ; j<strlen(word)-strlen(prefixes[i]) ; j++ )
				word[j] = word[j+strlen(prefixes[i])];
			word[j] = '\0';
			return;
		}
	}
}
#endif PREFIXES

 void strip_suffixes (char* word )
{
	step_1 ( word );
	step_2 ( word );
	step_3 ( word );
	step_4 ( word );
	step_5 ( word );
}

 void step_1 (char* word )
{
	char stem[BIG_KEYWORDSIZE];
	if ( word[strlen(word)-1] == 's' ) {
		if ( (has_suffix(word,"sses",stem) == TRUE) ||
				(has_suffix(word,"ies",stem) == TRUE ) )
			word[strlen(word)-2] = '\0';
		else
			if ( word[strlen(word)-2] != 's' )
				word[strlen(word)-1] = '\0';
		}
	if ( has_suffix(word,"eed",stem) == TRUE ) {
		if ( measure(stem) > 0 )
			word[strlen(word)-1] = '\0';
	}
	else {	if ( ( (has_suffix(word,"ed",stem) == TRUE )
				|| (has_suffix(word,"ing",stem) == TRUE ) )
				&& (contains_vowel(stem) == TRUE ) ) {
			word[strlen(stem)] = '\0';
			if ( ( has_suffix(word,"at",stem) == TRUE )
				|| ( has_suffix(word,"bl",stem) == TRUE )
				|| ( has_suffix(word,"iz",stem) == TRUE ) ) {
				word[strlen(word)+1] = '\0';
				word[strlen(word)] = 'e';
			}
			else {	int length = strlen(word);
				if ( (word[length-1] == word[length-2])
						&& (word[length-1] != 'l')
						&& (word[length-1] != 's')
						&& (word[length-1] != 'z') )
					word[length-1] = '\0';
				else
					if ( measure(word) == 1 ) {
					     if ( cvc(word) == TRUE ) {
						word[strlen(word)+1] = '\0';
						word[strlen(word)] = 'e';
						}
					}
			}
		}
	}
	if ( (has_suffix(word,"y",stem) == TRUE) &&
					(contains_vowel(stem) == TRUE) )
		word[strlen(word)-1] = 'i';
}

 void step_2 (char* word )
{
	 char *suffixes[][2] =  { { "ational", "ate" },
				        { "tional",  "tion" },
				        { "enci",    "ence" },
					{ "anci",    "ance" },
					{ "izer",    "ize" },
					{ "iser",    "ize" },
					{ "abli",    "able" },
					{ "alli",    "al" },
					{ "entli",   "ent" },
					{ "eli",     "e" },
					{ "ousli",   "ous" },
					{ "ization", "ize" },
					{ "isation", "ize" },
					{ "ation",   "ate" },
					{ "ator",    "ate" },
					{ "alism",   "al" },
					{ "iveness", "ive" },
					{ "fulness", "ful" },
					{ "ousness", "ous" },
					{ "aliti",   "al" },
					{ "iviti",   "ive" },
					{ "biliti",  "ble" },
					{  0,        0     } };
	char stem[BIG_KEYWORDSIZE];
	register int index;
	for ( index = 0 ; suffixes[index][0] != 0 ; index++ ) {
		if ( has_suffix ( word, suffixes[index][0], stem ) == TRUE ) {
			if ( measure ( stem ) > 0 ) {
				sprintf ( word, "%s%s", stem, suffixes[index][1] );
				return;
			}
		}
	}
	return;
}

 void step_3 (char* word )
{
	 char *suffixes[][2] = { { "icate", "ic" },
				       { "ative", "" },
				       { "alize", "al" },
				       { "alise", "al" },
				       { "iciti", "ic" },
				       { "ical",  "ic" },
				       { "ful",   "" },
				       { "ness",  "" },
				       { 0,       0 } };
	char stem[BIG_KEYWORDSIZE];
	register int index;
	for ( index = 0 ; suffixes[index][0] != 0 ; index++ ) {
		if ( has_suffix ( word, suffixes[index][0], stem ) == TRUE )
			if ( measure ( stem ) > 0 ) {
				sprintf ( word, "%s%s", stem, suffixes[index][1] );
				return;
			}
	}
	return;
}

 void step_4 (char* word )
{
	 char *suffixes[] = { "al", "ance", "ence", "er", "ic", "able",
		"ible", "ant", "ement", "ment", "ent", "sion", "tion",
		"ou", "ism", "ate", "iti", "ous", "ive", "ize", "ise", 0 };
	char stem[BIG_KEYWORDSIZE];
	register int index;
	for ( index = 0 ; suffixes[index] != 0 ; index++ ) {
		if ( has_suffix ( word, suffixes[index], stem ) == TRUE )
			if ( measure ( stem ) > 1 ) {
				strcpy( word, stem );
				return;
			}
	}
	return;
}

 void step_5 (char* word )
{
	if ( word[strlen(word)-1] == 'e' ) {
		if ( measure(word) > 1 )
			/* measure(word)==measure(stem) if ends in vowel */
			word[strlen(word)-1] = '\0';
		else
			if ( measure(word) == 1 ) {
				char stem[BIG_KEYWORDSIZE];
				strcpy(stem,"");
				strncat( stem, word, strlen(word)-1 );
				if ( cvc(stem) == FALSE )
					word[strlen(word)-1] = '\0';
			}
	}
	if ( (word[strlen(word)-1] == 'l')
				&& (word[strlen(word)-2] == 'l')
				&& (measure(word) > 1) )
		word[strlen(word)-1] = '\0';
}

 has_suffix (char* word,char* suffix,char* stem )
{
	char tmp[BIG_KEYWORDSIZE];
	if ( strlen(word) <= strlen(suffix) )
		return(FALSE);
	if ( (strlen(suffix) > 1) &&
			( word[strlen(word)-2] != suffix[strlen(suffix)-2] ) )
		return(FALSE);
	strcpy ( stem, "" );
	strncat ( stem, word, strlen(word) - strlen(suffix) );
	strcpy ( tmp, stem );
	strcat ( tmp, suffix );
	if ( strcmp (  tmp, word ) == 0 )
		return(TRUE);
	else
		return(FALSE);
}

 int cvc (char* word )
{
	int length=strlen(word);
	if ( length < 3 )
		return(FALSE);
	if ( (vowel(word[length-1],word[length-2]) == FALSE)
			&& (word[length-1] != 'w')
			&& (word[length-1] != 'x')
			&& (word[length-1] != 'y')
			&& (vowel(word[length-2],word[length-3]) == TRUE)
			&& ( ( (length==3) && (vowel(word[0],'a') == FALSE) )
			|| (vowel(word[length-3],word[length-4])==FALSE) ) )
		return(TRUE);
	else
		return(FALSE);
}

 int vowel (char ch,char prev )
{
	switch ( ch ) {
		case 'a':
		case 'e':
		case 'i':
		case 'o':
		case 'u': return(TRUE);
			  break;
		case 'y': return( vowel(prev,'?') == TRUE );
			  break;
		default : return(FALSE);
			  break;
	}
}

 measure (char* stem )
{
	register int i=0, count = 0;
	int length=strlen(stem);
	while ( i < length ) {
		for ( ; i < length ; i++ ) {
			if ( i > 0 ) {
				if ( vowel(stem[i],stem[i-1]) == TRUE )
					break;
			}
			else {	if ( vowel(stem[i],'a') == TRUE )
					break;
			}
		}
		for ( i++ ; i < length ; i++ ) {
			if ( i > 0 ) {
				if ( vowel(stem[i],stem[i-1]) == FALSE )
					break;
			}
			else {	if ( vowel(stem[i],'?') == FALSE )
					break;
			}
		}
		if ( i < length ) {
			count++;
			i++;
		}
	}
	return(count);
}

 int contains_vowel (char* word )
{
	register int i;
	for ( i=0 ; i < strlen(word) ; i++ )
		if ( i > 0 ) {
			if ( vowel(word[i],word[i-1]) == TRUE )
				return(TRUE);
		}
		else {	if ( vowel(word[0],'a') == TRUE )
				return(TRUE);
		}
	return(FALSE);
}

