#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ddi.h> // max
#include <errno.h>
#include <iostream.h>
#include "merge_inv_files.h"
#include <stdlib.h>
#include "../index/InvertedFile5.h"
#include "../index/Index5.h"
#include <math.h>

#define BATCH_SIZE 2000 // num tuples
#define MAX_DOCS_PER_BATCH 2000
#define MAX_DOCS_PER_OUTPAGE 2000
#define MAX_BUFFER_SIZE 10000

ofstream GLOBAL_logFileStream;

void get_curr(merge_struct *table_, int slot_, int &doc_id_, int &word_pos_) {
  doc_id_ = table_[slot_].array[table_[slot_].position].doc_id;
  word_pos_ = table_[slot_].array[table_[slot_].position].word_pos;
}

// return 1 if slot_ has no more values remaining
int go_next(merge_struct *table_, int slot_) {
  table_[slot_].position++;
  if(table_[slot_].position == table_[slot_].num_in_batch) {
    if(table_[slot_].num_tuples_left == 0) return 1;
    else {
      table_[slot_].num_in_batch = (MAX_DOCS_PER_BATCH < table_[slot_].num_tuples_left ? MAX_DOCS_PER_BATCH : table_[slot_].num_tuples_left);
      table_[slot_].position = 0;
      if(lseek(table_[slot_].file_fd, table_[slot_].file_offset, SEEK_SET) < 0) return 1;
      for(int i=0; i<table_[slot_].num_in_batch; i++) {
        if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].doc_id, sizeof(int)) < sizeof(int)) return 1;
        if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].word_pos, sizeof(int)) < sizeof(int)) return 1;
      }
      table_[slot_].file_offset += (table_[slot_].num_in_batch * 2 * sizeof(int));
      table_[slot_].num_tuples_left -= table_[slot_].num_in_batch;
      return 0;
    }
  }    
  else return 0;
}

int go_first(merge_struct *table_, int slot_, int file_fd_, int file_offset_, int num_tuples_) {
  table_[slot_].file_fd = file_fd_;
  table_[slot_].file_offset = file_offset_;
  table_[slot_].num_tuples_left = num_tuples_;

  table_[slot_].num_in_batch = (MAX_DOCS_PER_BATCH < table_[slot_].num_tuples_left ? 
				MAX_DOCS_PER_BATCH : table_[slot_].num_tuples_left);
  table_[slot_].position = 0;
  table_[slot_].array = new (doc_word_pos_struct)[MAX_DOCS_PER_BATCH];
  if(lseek(table_[slot_].file_fd, table_[slot_].file_offset, SEEK_SET) < 0) return 1;
  for(int i=0; i<table_[slot_].num_in_batch; i++) {
    if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].doc_id, sizeof(int)) < sizeof(int)) return 1;
    if(read(table_[slot_].file_fd, (char *)&table_[slot_].array[i].word_pos, sizeof(int)) < sizeof(int)) return 1;
  }
  table_[slot_].file_offset += (table_[slot_].num_in_batch * 2 * sizeof(int));
  table_[slot_].num_tuples_left -= table_[slot_].num_in_batch;
  if(table_[slot_].position == table_[slot_].num_in_batch) return 1;
  return 0;
}

void print_out_array(int size_, doc_word_pos_struct *file_ptr_, int qr_id) {
  
  for(int i=0; i<size_; i++) {
    if(write(qr_id, (char *)&(file_ptr_[i].doc_id), sizeof(int)) < sizeof(int)) return;
    if(write(qr_id, (char *)&(file_ptr_[i].word_pos), sizeof(int)) < sizeof(int)) return;    
  }
  
  //for(int i=0; i<size_; i++) {
  //cout << file_ptr_[i].doc_id << " " << file_ptr_[i].word_pos << endl;
  //}
}

void merge_lists(FileDescPtr first_ptr_, FileDescPtr second_ptr_, FileDescPtr result_ptr_) {
  int min_doc_id, min_word_pos, min_doc_slot, doc_id, word_pos, skip_count;
  int done_slot_count, i;
  doc_word_pos_struct* out_array = new (doc_word_pos_struct)[MAX_DOCS_PER_OUTPAGE];
  int num_lists = 2, out_docs_count;
  merge_struct* table = new (merge_struct) [num_lists];
  int *done_slots = new (int) [num_lists];

  lseek(result_ptr_.fd, result_ptr_.offset, SEEK_SET);

  done_slot_count = 0;
  done_slots[0] = go_first(table, 0, first_ptr_.fd, first_ptr_.offset, first_ptr_.ntuples);
  if(done_slots[0] == 1) done_slot_count++;
  done_slots[1] = go_first(table, 1, second_ptr_.fd, second_ptr_.offset, second_ptr_.ntuples);
  if(done_slots[1] == 1) done_slot_count++;  

  out_docs_count = 0;
  while(done_slot_count < num_lists) {
    skip_count = 0;
    for(i=0; i<num_lists; i++) {
      if(done_slots[i] == 1) {
	skip_count++;
	continue;
      }

      get_curr(table, i, doc_id, word_pos);
      if(skip_count == i || doc_id < min_doc_id || (doc_id == min_doc_id && word_pos < min_word_pos)) {
	min_doc_id = doc_id;
	min_word_pos = word_pos;
	min_doc_slot = i;
      }
    }
    
    
    if(go_next(table, min_doc_slot)) {
      if(done_slots[min_doc_slot] == 0) {
	done_slots[min_doc_slot] = 1;
	done_slot_count++;
      }
    }

    // output min_doc_id && min_word_pos
    out_array[out_docs_count%MAX_DOCS_PER_OUTPAGE].doc_id = min_doc_id;
    out_array[out_docs_count%MAX_DOCS_PER_OUTPAGE].word_pos  = min_word_pos;
    out_docs_count++;
    if(out_docs_count%MAX_DOCS_PER_OUTPAGE == 0) {
      // print the out array
      print_out_array(MAX_DOCS_PER_OUTPAGE, out_array, result_ptr_.fd);
    }
  }

  if(out_docs_count%MAX_DOCS_PER_OUTPAGE > 0) {
    // print the out array
    print_out_array(out_docs_count%MAX_DOCS_PER_OUTPAGE, out_array, result_ptr_.fd);
  }

  for(i=0; i<num_lists; i++) 
    delete [](table[i].array);
  delete []table;
  delete []out_array;
  delete []done_slots;
  return;
}

void merge_inv_lists(File_Ptr &first_, File_Ptr &second_, File_Ptr &result_) {
  int first_fd, second_fd, result_fd, i, j, condition;
  int first_npaths, second_npaths, result_npaths, end_offset, first_path_len, second_path_len;
  int first_ntuples, second_ntuples, num_tuples, result_ntuples, record_end_offset, num_tuples_offset;  
  bool read_first, read_second;
  char *first_buffer, *second_buffer, *first_path, *second_path;
  FileDescPtr first_ptr, second_ptr, result_ptr;
  
  first_buffer = new char[2 * sizeof(int) * BATCH_SIZE];
  second_buffer = new char[2 * sizeof(int) * BATCH_SIZE];

  first_fd = first_.fd;
  if(lseek(first_fd, first_.offset, SEEK_SET) < 0) return;

  second_fd = second_.fd;
  if(lseek(second_fd, second_.offset, SEEK_SET) < 0) return;

  result_fd = result_.fd;
  
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;
  
  if(read(first_fd, (char *)&first_npaths, sizeof(int)) < sizeof(int)) return;
  if(read(second_fd, (char *)&second_npaths, sizeof(int)) < sizeof(int)) return;
  result_npaths = 0;
  if(write(result_fd, (char *)&result_npaths, sizeof(int)) < sizeof(int)) return;

  read_first = true;
  read_second = true;
  i = 0; j = 0;
  while(i < first_npaths || j < second_npaths) {
    result_npaths++;

    if(i < first_npaths && j < second_npaths) {
      if(read_first) {
	if(read(first_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(first_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	
	first_path = new char[first_path_len + 1];
	if(read(first_fd, (char *)first_path, first_path_len) < first_path_len) return;
	first_path[first_path_len] = 0;
	read_first = false;

      }
      if(read_second) {
	if(read(second_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(second_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	
	second_path = new char[second_path_len + 1];
	if(read(second_fd, (char *)second_path, second_path_len) < second_path_len) return;
	second_path[second_path_len] = 0;
	read_second = false;
      }

      if((condition = strcmp(first_path, second_path)) < 0) {
	if(write(result_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, first_path, first_path_len) < first_path_len) return;

	num_tuples = first_ntuples;
	while(num_tuples > BATCH_SIZE) {
	  if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  num_tuples -= BATCH_SIZE;
	}

	if(num_tuples > 0) {
	  if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	  if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	}

	read_first = true;
	delete []first_path;
	i++;
      }
      else if(condition > 0) {
	if(write(result_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, second_path, second_path_len) < second_path_len) return;

	num_tuples = second_ntuples;
	while(num_tuples > BATCH_SIZE) {
	  if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	  num_tuples -= BATCH_SIZE;
	}
	
	if(num_tuples > 0) {
	  if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	  if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	}

	read_second = true;
	delete []second_path;
	j++;
      }
      else {
	if(write(result_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	num_tuples_offset = lseek(result_fd, 0, SEEK_CUR);
	if(write(result_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	if(write(result_fd, second_path, second_path_len) < second_path_len) return;
		
	first_ptr.fd = first_fd;
	first_ptr.offset = lseek(first_fd, 0, SEEK_CUR);
	first_ptr.ntuples = first_ntuples;

	second_ptr.fd = second_fd;
	second_ptr.offset = lseek(second_fd, 0, SEEK_CUR);
	second_ptr.ntuples = second_ntuples;

	result_ptr.fd = result_fd;
	result_ptr.offset = num_tuples_offset + sizeof(int) + second_path_len;
	result_ptr.ntuples = 0;

	merge_lists(first_ptr, second_ptr, result_ptr);

	record_end_offset = lseek(result_fd, 0, SEEK_CUR);

	if(lseek(result_fd, num_tuples_offset, SEEK_SET) < 0) return;
	result_ntuples = (record_end_offset - result_ptr.offset)/ (2 * sizeof(int));
	if(write(result_fd, (char *)&result_ntuples, sizeof(int)) < sizeof(int)) return;
	if(lseek(result_fd, record_end_offset, SEEK_SET) < 0) return;	
	
	read_first = true;
	read_second = true;
	delete []first_path;
	delete []second_path;	
	i++;
	j++;
      }
    }
    else if(i < first_npaths) {
      if(read_first) {
	if(read(first_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(first_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
	
	first_path = new char[first_path_len + 1];
	if(read(first_fd, (char *)first_path, first_path_len) < first_path_len) return;
	first_path[first_path_len] = 0;
      }      
      
      if(write(result_fd, (char *)&first_path_len, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, (char *)&first_ntuples, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, first_path, first_path_len) < first_path_len) return;

      num_tuples = first_ntuples;
      while(num_tuples > BATCH_SIZE) {
	if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	num_tuples -= BATCH_SIZE;
      }

      if(num_tuples > 0) {
	if(read(first_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	if(write(result_fd, (char *)first_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
      }

      read_first = true;
      delete []first_path;
      i++;
    }
    else if(j < second_npaths) {
      if(read_second) {
	if(read(second_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
	if(read(second_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
	
	second_path = new char[second_path_len + 1];
	if(read(second_fd, (char *)second_path, second_path_len) < second_path_len) return;
	second_path[second_path_len] = 0;
      }

      if(write(result_fd, (char *)&second_path_len, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, (char *)&second_ntuples, sizeof(int)) < sizeof(int)) return;
      if(write(result_fd, second_path, second_path_len) < second_path_len) return;

      num_tuples = second_ntuples;
      while(num_tuples > BATCH_SIZE) {
	if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * BATCH_SIZE)) < (2 * sizeof(int) * BATCH_SIZE)) return;
	num_tuples -= BATCH_SIZE;
      }
      if(num_tuples > 0) {
	if(read(second_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
	if(write(result_fd, (char *)second_buffer, (2 * sizeof(int) * num_tuples)) < (2 * sizeof(int) * num_tuples)) return;
      }

      read_second = true;
      delete []second_path;
      j++;
    }
  }

  end_offset = lseek(result_fd, 0, SEEK_CUR);
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;
  if(write(result_fd, (char *)&result_npaths, sizeof(int)) < sizeof(int)) return;
  result_.offset = end_offset;
  if(lseek(result_fd, result_.offset, SEEK_SET) < 0) return;

  first_.offset = lseek(first_fd, 0, SEEK_CUR);
  second_.offset = lseek(second_fd, 0, SEEK_CUR);

  delete []first_buffer;
  delete []second_buffer;
}


// args:
// 1 & 3: files to be merged
// 2 & 4: level numbers of the index files [1] & [3] respectively
//        level number of 0 implies records of varying length
// 5    : output file name. Its level number (by construction) is 0
int main(int argc, char* argv[]) {
  struct timeval startTime;
  struct timeval stopTime;
  int first_fd, second_fd, result_fd, condition, result_listlen_offset;
  int first_rec_size, second_rec_size, second_offset, first_offset, first_eof, second_eof;
  int first_keylen, second_keylen, first_listlen, second_listlen, result_listlen;
  char *first_key, *second_key, *first_buffer, *second_buffer;
  struct stat buf;
  File_Ptr first_desc, second_desc, result_desc;
  bool read_first, read_second;
  FileHeader result_fh;
  GLOBAL_logFileStream.open("merge_inv_files.log", ios::out | ios::app);
  GLOBAL_logFileStream.setf(ios::fixed, ios::floatfield);
  RecHeader first_rh, second_rh, result_rh;

  if(argc < 6) {
    cout << "Usage: " << argv[0] << " <inv file 1> <lnum of file 1> <inv file 2> <lnum of file 2> <out inv file>" << endl;
    cout << "       " << "(lnums are usually 0)" << endl;
    exit(0);
  }
  GLOBAL_logFileStream << "Merging " << argv[1] << " and " << argv[3] << endl;

  first_buffer = new char[MAX_BUFFER_SIZE];
  second_buffer = new char[MAX_BUFFER_SIZE];

  first_fd = open(argv[1], O_RDONLY);
  if(first_fd < 0) return 0;

  first_rec_size = atoi(argv[2]);
  if(first_rec_size > 0) first_rec_size =  BASIC_REC_SIZE * (int) pow(3, first_rec_size - 1);

  second_fd = open(argv[3], O_RDONLY);
  if(second_fd < 0) return 0;

  second_rec_size = atoi(argv[4]);
  if(second_rec_size > 0) second_rec_size =  BASIC_REC_SIZE * (int) pow(3, second_rec_size - 1);

  if(stat(argv[5], &buf) && errno == ENOENT)
    result_fd = open(argv[5], O_CREAT |  O_WRONLY, S_IREAD | S_IWRITE);
  else {
    cout << argv[5] << " exists already" << endl;
    exit(0);
  }
 
  assert(gettimeofday(&startTime, NULL) != -1);

  first_eof = lseek(first_fd, 0, SEEK_END);
  first_offset = sizeof(FileHeader);
  if(lseek(first_fd, first_offset, SEEK_SET) < 0) return 0;

  second_eof = lseek(second_fd, 0, SEEK_END);
  second_offset = sizeof(FileHeader);
  if(lseek(second_fd, second_offset, SEEK_SET) < 0) return 0;

  if(write(result_fd, (char *)&result_fh, sizeof(FileHeader)) < sizeof(FileHeader)) return 0;
  
  read_first = true;
  read_second = true;
  while(first_offset < first_eof || second_offset < second_eof) {
    if(first_offset < first_eof && second_offset < second_eof) {
      if(read_first) {
	if(read(first_fd, (char *)&first_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
	if(read(first_fd, (char *)&first_keylen, sizeof(int)) < sizeof(int)) return 0;
	first_key = new char[first_keylen + 1];
	if(read(first_fd, first_key, first_keylen) < first_keylen) return 0;
	first_key[first_keylen] = 0;
	if(read(first_fd, (char *)&first_listlen, sizeof(int)) < sizeof(int)) return 0;
	read_first = false;
      }
      if(read_second) {
	if(read(second_fd, (char *)&second_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
	if(read(second_fd, (char *)&second_keylen, sizeof(int)) < sizeof(int)) return 0;
	second_key = new char[second_keylen + 1];
	if(read(second_fd, second_key, second_keylen) < second_keylen) return 0;
	second_key[second_keylen] = 0;
	if(read(second_fd, (char *)&second_listlen, sizeof(int)) < sizeof(int)) return 0;
	read_second = false;
      }
      
      condition = strcmp(first_key, second_key);
      if(condition < 0) {
	if(write(result_fd, (char *)&result_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
	if(write(result_fd, (char *)&first_keylen, sizeof(int)) < sizeof(int)) return 0;
	if(write(result_fd, (char *)first_key, first_keylen) < first_keylen) return 0;
	result_listlen = first_listlen;
	if(write(result_fd, (char *)&result_listlen, sizeof(int)) < sizeof(int)) return 0;
	
	while(result_listlen > MAX_BUFFER_SIZE) {
	  if(read(first_fd, first_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	  if(write(result_fd, first_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	  result_listlen -= MAX_BUFFER_SIZE;
	}
	if(result_listlen > 0) {
	  if(read(first_fd, first_buffer, result_listlen) < result_listlen) return 0;
	  if(write(result_fd, first_buffer, result_listlen) < result_listlen) return 0;
	}
	
	read_first = true;
	first_offset += (sizeof(RecHeader) + (first_rec_size > 0 ? first_rec_size : 2*sizeof(int) + first_keylen + first_listlen));
	if(lseek(first_fd, first_offset, SEEK_SET) , 0) return 0;
      }
      else if(condition > 0) {
	if(write(result_fd, (char *)&result_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
	if(write(result_fd, (char *)&second_keylen, sizeof(int)) < sizeof(int)) return 0;
	if(write(result_fd, (char *)second_key, second_keylen) < second_keylen) return 0;
	result_listlen = second_listlen;
	if(write(result_fd, (char *)&result_listlen, sizeof(int)) < sizeof(int)) return 0;
	
	while(result_listlen > MAX_BUFFER_SIZE) {
	  if(read(second_fd, second_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	  if(write(result_fd, second_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	  result_listlen -= MAX_BUFFER_SIZE;
	}
	if(result_listlen > 0) {
	  if(read(second_fd, second_buffer, result_listlen) < result_listlen) return 0;
	  if(write(result_fd, second_buffer, result_listlen) < result_listlen) return 0;
	}
	
	read_second = true;
	second_offset += (sizeof(RecHeader) + (second_rec_size > 0 ? second_rec_size : 2*sizeof(int) + second_keylen + second_listlen));
	if(lseek(second_fd, second_offset, SEEK_SET) , 0) return 0;
      }
      else {
	if(write(result_fd, (char *)&result_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
	if(write(result_fd, (char *)&second_keylen, sizeof(int)) < sizeof(int)) return 0;
	if(write(result_fd, (char *)second_key, second_keylen) < second_keylen) return 0;
	result_listlen = 0;
	result_listlen_offset = lseek(result_fd, 0, SEEK_CUR);
	if(write(result_fd, (char *)&result_listlen, sizeof(int)) < sizeof(int)) return 0;

	first_desc.fd = first_fd;
	first_desc.offset = lseek(first_fd, 0, SEEK_CUR);
	second_desc.fd = second_fd;
	second_desc.offset = lseek(second_fd, 0, SEEK_CUR);
	result_desc.fd = result_fd;
	result_desc.offset = lseek(result_fd, 0, SEEK_CUR);

	merge_inv_lists(first_desc, second_desc, result_desc);

	result_listlen = result_desc.offset-result_listlen_offset-sizeof(int);
	if(lseek(result_fd, result_listlen_offset, SEEK_SET) < 0) return 0;
	if(write(result_fd, (char *)&result_listlen, sizeof(int)) < sizeof(int)) return 0;
	if(lseek(result_fd, result_desc.offset, SEEK_SET) < 0) return 0;

	read_first = true;
	read_second = true;
	
	//cout << "r.offset = " << result_desc.offset << " f.offset = " << first_desc.offset << " s.offset = " << second_desc.offset << endl;
	first_offset += (sizeof(RecHeader) + (first_rec_size > 0 ? first_rec_size : 2*sizeof(int) + first_keylen + first_listlen));	
	second_offset += (sizeof(RecHeader) + (second_rec_size > 0 ? second_rec_size : 2*sizeof(int) + second_keylen + second_listlen));
	if(lseek(first_fd, first_offset, SEEK_SET) , 0) return 0;
	if(lseek(second_fd, second_offset, SEEK_SET) , 0) return 0;
      }
    }
    else if(first_offset < first_eof) {
      if(read_first) {
	if(read(first_fd, (char *)&first_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
	if(read(first_fd, (char *)&first_keylen, sizeof(int)) < sizeof(int)) return 0;
	first_key = new char[first_keylen + 1];
	if(read(first_fd, first_key, first_keylen) < first_keylen) return 0;
	first_key[first_keylen] = 0;
	if(read(first_fd, (char *)&first_listlen, sizeof(int)) < sizeof(int)) return 0;
      }

      if(write(result_fd, (char *)&result_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
      if(write(result_fd, (char *)&first_keylen, sizeof(int)) < sizeof(int)) return 0;
      if(write(result_fd, (char *)first_key, first_keylen) < first_keylen) return 0;
      result_listlen = first_listlen;
      if(write(result_fd, (char *)&result_listlen, sizeof(int)) < sizeof(int)) return 0;

      while(result_listlen > MAX_BUFFER_SIZE) {
	if(read(first_fd, first_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	if(write(result_fd, first_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	result_listlen -= MAX_BUFFER_SIZE;
      }
      if(result_listlen > 0) {
	if(read(first_fd, first_buffer, result_listlen) < result_listlen) return 0;
	if(write(result_fd, first_buffer, result_listlen) < result_listlen) return 0;
      }

      read_first = true;
      first_offset += (sizeof(RecHeader) + (first_rec_size > 0 ? first_rec_size : 2*sizeof(int) + first_keylen + first_listlen));
      if(lseek(first_fd, first_offset, SEEK_SET) , 0) return 0;
    }
    else {
      if(read_second) {
	if(read(second_fd, (char *)&second_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
	if(read(second_fd, (char *)&second_keylen, sizeof(int)) < sizeof(int)) return 0;
	second_key = new char[second_keylen + 1];
	if(read(second_fd, second_key, second_keylen) < second_keylen) return 0;
	second_key[second_keylen] = 0;
	if(read(second_fd, (char *)&second_listlen, sizeof(int)) < sizeof(int)) return 0;
	read_second = false;
      }

      if(write(result_fd, (char *)&result_rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
      if(write(result_fd, (char *)&second_keylen, sizeof(int)) < sizeof(int)) return 0;
      if(write(result_fd, (char *)second_key, second_keylen) < second_keylen) return 0;
      result_listlen = second_listlen;
      if(write(result_fd, (char *)&result_listlen, sizeof(int)) < sizeof(int)) return 0;

      while(result_listlen > MAX_BUFFER_SIZE) {
	if(read(second_fd, second_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	if(write(result_fd, second_buffer, MAX_BUFFER_SIZE) < MAX_BUFFER_SIZE) return 0;
	result_listlen -= MAX_BUFFER_SIZE;
      }
      if(result_listlen > 0) {
	if(read(second_fd, second_buffer, result_listlen) < result_listlen) return 0;
	if(write(result_fd, second_buffer, result_listlen) < result_listlen) return 0;
      }

      read_second = true;
      second_offset += (sizeof(RecHeader) + (second_rec_size ? second_rec_size : 2*sizeof(int) + second_keylen + second_listlen));
      if(lseek(second_fd, second_offset, SEEK_SET) , 0) return 0;
    }
  }
  assert(gettimeofday(&stopTime, NULL) != -1);
  double mergeTime = 
    (stopTime.tv_sec-startTime.tv_sec) + 
    (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
  GLOBAL_logFileStream << "Merge Time = " << mergeTime << " seconds" << endl << endl;

  delete []first_buffer;
  delete []second_buffer;
  close(first_fd);
  close(second_fd);
  close(result_fd);
  GLOBAL_logFileStream.close();
}

